from __future__ import annotations
from boto3 import client as BotoClient
from botocore.client import BaseClient as BotoEcs
from collections import namedtuple
from datetime import datetime, timezone
from functools import lru_cache
import io
import json
import os
import requests
import sys
from termcolor import colored
from typing import List, Literal, Optional, Tuple, Union
from dcicutils.datetime_utils import format_datetime
from dcicutils.misc_utils import format_size
from dcicutils.secrets_utils import get_identity_secrets
from hms_utils.aws.codebuild.utils import get_image_build_info
from hms_utils.chars import chars
from hms_utils.datetime_utils import convert_uptime_to_datetime, format_duration
from hms_utils.threading_utils import run_concurrently
from hms_utils.version_utils import get_version


class AwsEcs:

    BLUE = "blue"
    GREEN = "green"
    BLUE_OR_GREEN = [BLUE, GREEN]

    PORTAL = "Portal"
    INDEXER = "Indexer"
    INGESTER = "Ingester"
    TYPES = [PORTAL, INDEXER, INGESTER]

    class Cluster:
        def __init__(self, cluster_arn: str, ecs: Optional[AwsEcs] = None) -> None:
            self.cluster_arn = cluster_arn or ""
            self.cluster_name = AwsEcs._nonarn_name(cluster_arn)
            self._services = None
            self._ecs = ecs if isinstance(ecs, AwsEcs) else AwsEcs()
            self._ecs._note_name(self.cluster_name)
        @property  # noqa
        def services(self) -> List[AwsEcs.Service]:
            if self._services is None:
                self._services = []
                if service_names := self._ecs._list_services(cluster_name=self.cluster_name):
                    if service_descriptions := self._ecs._describe_services(cluster_name=self.cluster_name,
                                                                            service_names=service_names):
                        for service_description in service_descriptions:
                            self._services.append(AwsEcs.Service(
                                cluster=self,
                                service_name=service_description.get("serviceName"),
                                service_arn=service_description.get("serviceArn"),
                                task_definition=service_description.get("taskDefinition"),
                                ecs=self._ecs))
                    self._services = self._sort_services_by_type(self._services)
            return self._services
        @property  # noqa
        def running_tasks(self) -> List[object]:  # noqa
            return self._ecs._list_running_tasks(cluster_name=self.cluster_name)
        @property  # noqa
        def is_blue(self) -> bool:
            return AwsEcs._is_blue(self.cluster_name)
        @property  # noqa
        def is_green(self) -> bool:
            return AwsEcs._is_green(self.cluster_name)
        @property  # noqa
        def blue_or_green(self) -> str:
            return AwsEcs._blue_or_green(self.cluster_name)
        @property  # noqa
        def is_mirrored(self) -> Optional[bool]:
            if not self.blue_or_green:
                return None
            for service in self.services:
                if not service.is_mirrored:
                    return False
            return True
        @property  # noqa
        def annotation(self) -> str:
            if blue_or_green := self.blue_or_green:
                return self._ecs._terminal_color(blue_or_green.upper(), blue_or_green)
            return ""
        @staticmethod  # noqa
        def _sort_services_by_type(services: List[AwsEcs.Service]) -> List[AwsEcs.Service]:
            sorted_services = []
            for service_type in AwsEcs.TYPES:
                for service in services:
                    if service.type == service_type:
                        sorted_services.append(service)
            return sorted_services

        def print_cluster(self, shortened_names: bool = False, versioned_names: bool = False,
                          nodns: bool = False, nocontainer: bool = False, noimage: bool = False,
                          nogit: bool = False, notasks: bool = False, nohealth: bool = False, nouptime: bool = False,
                          show: bool = False, verbose: bool = False, noprint: bool = False) -> Optional[str]:

            cluster = self
            cluster_running_task_count = len(cluster.running_tasks) if (not notasks) else 0
            cluster_is_data = False
            lines = [] ; container_lines_previous = None  # noqa
            cluster_is_mirrored_state = True if cluster.blue_or_green else None
            if services := cluster.services:
                cluster_annotation = cluster.annotation
                line = (f"\n{chars.rarrow} CLUSTER:"
                        f" {self._ecs.format_name(cluster.cluster_name, shortened=shortened_names)}"
                        f"__REPLACE_STAGING_DATA__"
                        f"{f' | {cluster_annotation}' if cluster_annotation else ''}__REPLACEBELOW__")  # noqa
                if not notasks:
                    line += f"{f' | ({cluster_running_task_count})' if cluster_running_task_count > 0 else ''}"
                lines.append(line)
                cluster_line_index = len(lines) - 1
                service_running_task_total_count = 0
                for service in services:
                    if cluster.blue_or_green and (not service.is_mirrored):
                        cluster_is_mirrored_state = False
                    service_aname = (service.dns_aname
                                     if (not nodns) and AwsEcs._is_portal(service.service_name) else None)
                    service_cname = (service.dns_cname
                                     if (not nodns) and AwsEcs._is_portal(service.service_name) else None)
                    if (not nohealth) and service_aname:
                        health = AwsEcs._get_health_page(service_cname or service_aname)
                    else:
                        health = None
                    service_running_task_count = len(service.running_tasks) if (not notasks) else 0
                    service_running_task_total_count += service_running_task_count
                    service_mirror_indicator = chars.square_hollow if service.is_mirrored else '-'
                    service_name = self._ecs.format_name(service.service_name,
                                                         shortened=shortened_names, versioned=versioned_names)
                    service_annotation = service.get_annotation(dns=not nodns)
                    # TODO: Hacky
                    if " STAGING " in service_annotation:
                        lines[cluster_line_index] = \
                            lines[cluster_line_index].replace("__REPLACE_STAGING_DATA__", " | STAGING")
                    elif " DATA " in service_annotation:
                        cluster_is_data = True
                        lines[cluster_line_index] = \
                            lines[cluster_line_index].replace("__REPLACE_STAGING_DATA__", " | DATA")
                    else:
                        lines[cluster_line_index] = \
                            lines[cluster_line_index].replace("__REPLACE_STAGING_DATA__", "")
                    task_definition_name = self._ecs.format_name(service.task_definition.task_definition_name,
                                                                 shortened=shortened_names, versioned=versioned_names)
                    task_definition_annotation = service.task_definition.annotation
                    lines.append(f"  {service_mirror_indicator} SERVICE: {service_name}"
                                 f"{f' | {service_annotation}' if service_annotation else ''}")
                    if service_aname:
                        if service_cname and cluster.blue_or_green:
                            service_cname = self._ecs._terminal_color(service_cname,
                                                                      service.task_definition.blue_or_green,
                                                                      underline=False)
                        line = (f"        DNS: {service_aname}"
                                f"{f' {chars.rarrow} {service_cname}' if service_cname else ''}")
                        if health:
                            if portal_version := health.get("project_version"):
                                line += f" {chars.dot_hollow} {portal_version}"
                            line += f" {chars.check if health else chars.xmark}"
                            if health_blue_or_green := AwsEcs._get_blue_or_green_from_health(health):
                                if ((cluster_is_mirrored_state and (cluster.blue_or_green != health_blue_or_green)) or
                                    ((not cluster_is_mirrored_state) and
                                     (cluster.blue_or_green == health_blue_or_green))):
                                    line += chars.check
                                else:
                                    line += chars.xmark
                                # line += f" ({health_blue_or_green})"
                        lines.append(line)
                        if verbose and (certificate_expiration_date := service.certificate_expiration_date):
                            certificate_expiration_duration = (
                                format_duration(certificate_expiration_date - datetime.now(timezone.utc), verbose=True))
                            lines.append(f"     CERTEX: {format_datetime(certificate_expiration_date)}"
                                         f" {chars.dot_hollow} {certificate_expiration_duration}")
                        if (not nouptime) and health and (portal_uptime := health.get("uptime")):
                            portal_uptime = convert_uptime_to_datetime(portal_uptime)
                            portal_uptime = format_duration(datetime.now(timezone.utc) - portal_uptime, verbose=True)
                            lines.append(f"     UPTIME: {portal_uptime}")
                    line = (
                        f"    -- TASK: {task_definition_name}"
                        f"{f' | {task_definition_annotation}' if task_definition_annotation else ''}")
                    if not notasks:
                        line += f"{f' | ({service_running_task_count})' if service_running_task_count > 0 else ''}"
                    lines.append(line)
                    if verbose and (target_group_arn := service.target_group_arn):
                        lines.append(f"       LBTG: {target_group_arn}")
                if cluster_is_mirrored_state is True:
                    lines[cluster_line_index] = lines[cluster_line_index].replace("__REPLACEBELOW__", " | MIRRORED")
                elif cluster_is_mirrored_state is False:
                    lines[cluster_line_index] = lines[cluster_line_index].replace("__REPLACEBELOW__", " | NORMAL")
                else:
                    lines[cluster_line_index] = lines[cluster_line_index].replace("__REPLACEBELOW__", "")
                if not notasks:
                    if cluster_running_task_count == service_running_task_total_count:
                        lines[cluster_line_index] += f" {chars.check}"
                    else:
                        lines[cluster_line_index] += f" {chars.xmark}"
            for service in services:
                if ((not nocontainer) and
                    (container := service.task_definition.get_container(service, noimage=noimage, nogit=nogit))):  # noqa
                    # The container info (for our purposes) is virtually always the same across services,
                    # i.e. portal, indexer, and ingester, within the clustser; so we only show this once
                    # per cluster, if all the info is exactly the same; we don't print container["name"] here
                    # because that actually is not unique and that would mess this up for our common case.
                    container_lines = []
                    container_lines.append(f"   IDENTITY: {container['identity']}")
                    if global_env_bucket := container.get("global_env_bucket"):
                        container_lines.append(f"        GEB: {global_env_bucket}")
                        if environment_name := container.get("environment_name"):
                            container_lines[len(container_lines) - 1] += f" {chars.dot_hollow} {environment_name}"
                        if (cluster.blue_or_green and
                            (s3 := AwsS3(global_env_bucket)) and
                            (main_ecosystem := s3.load_json_file("main.ecosystem"))):  # noqa
                            if ecosystem := main_ecosystem.get("ecosystem"):
                                container_lines.append(
                                    f"        ECO: ecosystem: {ecosystem.replace('.ecosystem', '')}")
                                if ecosystem := s3.load_json_file(ecosystem):
                                    if production_env_name := ecosystem.get("prd_env_name"):
                                        if staging_env_name := ecosystem.get("stg_env_name"):  # noqa
                                            container_lines[len(container_lines) - 1] += (
                                                f" {chars.dot_hollow} staging: {staging_env_name}"
                                                f" {chars.dot_hollow} data: {production_env_name}")
                                            if ((cluster_is_mirrored_state and
                                                 self._ecs._is_blue(production_env_name) and
                                                 self._ecs._is_green(staging_env_name)) or
                                                ((not cluster_is_mirrored_state) and
                                                 self._ecs._is_green(production_env_name) and
                                                 self._ecs._is_blue(staging_env_name))):
                                                container_lines[len(container_lines) - 1] += f" {chars.check}"
                                            else:
                                                container_lines[len(container_lines) - 1] += f" {chars.xmark}"
                                        else:
                                            container_lines[len(container_lines) - 1] += (
                                                f" {chars.dot_hollow} production: {production_env_name}")
                            elif production_env_name := main_ecosystem.get("prd_env_name"):
                                if staging_env_name := main_ecosystem.get("stg_env_name"):
                                    container_lines.append(
                                        f"        ECO: staging: {staging_env_name}"
                                        f" {chars.dot_hollow} data: {production_env_name}")
                                    if ((cluster_is_mirrored_state and
                                         self._ecs._is_blue(production_env_name) and
                                         self._ecs._is_green(staging_env_name)) or
                                        ((not cluster_is_mirrored_state) and
                                         self._ecs._is_green(production_env_name) and
                                         self._ecs._is_blue(staging_env_name))):
                                        container_lines[len(container_lines) - 1] += f" {chars.check}"
                                    else:
                                        container_lines[len(container_lines) - 1] += f" {chars.xmark}"
                                else:
                                    container_lines.append(
                                        f"        ECO: production: {production_env_name}")
                    if image := container.get("image"):
                        build_project = container.get("build_project")
                        image_pushed_at = container.get("image_pushed_at")
                        image_size = container.get("image_size")
                        container_lines.append(f"      IMAGE: {image}"
                                               f" ({format_size(image_size)}) {chars.dot_hollow}"
                                               f" {build_project} {chars.dot_hollow} {image_pushed_at}")
                        if verbose:
                            if (image_digest := container.get("image_digest")):
                                container_lines.append(f"     DIGEST: {image_digest}")
                        if git_repo := container.get("git_repo"):
                            git_branch = container.get("git_branch")
                            git_commit = container.get("git_commit")
                            container_lines.append(f"        GIT: {git_repo} {chars.dot_hollow} {git_branch}"
                                                   f" {chars.dot_hollow} {git_commit}")
                            if git_commit_date := container.get("git_commit_date"):
                                container_lines[len(container_lines) - 1] += (
                                    f" {chars.dot_hollow} {format_datetime(git_commit_date, notz=True)}")
                            if git_commit_latest := container.get("git_commit_latest"):
                                if git_commit_latest == git_commit:
                                    # Check mark to indicate this is the latest commit.
                                    container_lines[len(container_lines) - 1] += f" {chars.check}"
                                    if cluster_is_data:
                                        # Double check mark to indicate we don't need an identity-swap.
                                        container_lines[len(container_lines) - 1] += f"{chars.check}"
                                else:
                                    # Note that this image not built from the latest commit.
                                    # Find out how old this is compared to the latest.
                                    container_lines[len(container_lines) - 1] += f" {chars.xmark}"
                                    if git_commit_latest_date := container.get("git_commit_latest_date"):
                                        delta_seconds = (datetime.fromisoformat(git_commit_latest_date) -
                                                         datetime.fromisoformat(git_commit_date)).total_seconds()
                                        if delta_seconds > 0:
                                            if (delta_hours := int(delta_seconds / 3600)) > 0:
                                                if (delta_days := int(delta_hours / 24)) > 0:
                                                    container_lines[len(container_lines) - 1] += f" (-{delta_days}d)"
                                                else:
                                                    container_lines[len(container_lines) - 1] += f" (-{delta_hours}h)"
                    if elasticsearch_server := container.get("elasticsearch"):
                        elasticsearch_server = self._ecs._unversioned_name(elasticsearch_server)
                        if (cluster.blue_or_green and
                            (elasticsearch_blue_or_green := self._ecs._blue_or_green(elasticsearch_server))):  # noqa
                            elasticsearch_server = self._ecs._terminal_color(elasticsearch_server,
                                                                             elasticsearch_blue_or_green,
                                                                             underline=False)
                        container_lines.append(f"         ES: {elasticsearch_server}")
                    if database_server := container.get("database"):
                        container_lines.append(f"        RDS: {database_server}")
                    if redis_server := container.get("redis"):
                        container_lines.append(f"      REDIS: {redis_server}")
                    if s3_encrypt_key := container.get("s3_encrypt_key"):
                        if show:
                            container_lines.append(f"        SEK: {s3_encrypt_key}")
                        else:
                            container_lines.append(f"        SEK: {s3_encrypt_key[:2]}******")
                        if s3_encrypt_key_id := container.get("s3_encrypt_key_id"):
                            container_lines[len(container_lines) - 1] += (
                                f" {chars.dot_hollow} {s3_encrypt_key_id}")
                    elif s3_encrypt_key_id := container.get("s3_encrypt_key_id"):
                        container_lines.append(f"      SEKID: {s3_encrypt_key_id}")
                    if verbose and (auth0_client := container.get("auth0_client")):
                        container_lines.append(f"      AUTH0: {auth0_client}")
                        if auth0_secret := container.get("auth0_secret"):
                            if show:
                                container_lines[len(container_lines) - 1] += (
                                    f" {chars.dot_hollow} {auth0_secret}")
                            else:
                                container_lines[len(container_lines) - 1] += (
                                    f" {chars.dot_hollow} {auth0_secret[:2]}******")
                        if auth0_domain := container.get("auth0_domain"):
                            container_lines[len(container_lines) - 1] += (
                                f" {chars.dot_hollow} {auth0_domain}")
                    if container_lines != container_lines_previous:
                        lines += container_lines
                    container_lines_previous = container_lines
            if noprint is not True:
                for line in lines:
                    print(line)
            return None if (noprint is not True) else lines

    class Service:
        def __init__(self, cluster: AwsEcs.Cluster, service_name: str, service_arn: str,
                     task_definition: Optional[Union[AwsEcs.TaskDefinition, str]] = None,
                     ecs: Optional[AwsEcs] = None) -> None:
            self.cluster = cluster if isinstance(cluster, AwsEcs.Cluster) else None
            self.service_name = service_name or ""
            self.service_arn = service_arn or ""
            self.task_definition = (task_definition if isinstance(task_definition, AwsEcs.TaskDefinition)
                                    else (AwsEcs.TaskDefinition(task_definition, ecs=ecs)
                                          if isinstance(task_definition, str) else None))
            self._dns_aname = None
            self._dns_cname = None
            self._target_group_arn = None
            self._certificate_expiration_date = None
            self._ecs = ecs if isinstance(ecs, AwsEcs) else AwsEcs()
            self._ecs._note_name(self.service_name)
        @property  # noqa
        def running_tasks(self) -> List[object]:  # noqa
            return self._ecs._list_running_tasks(cluster_name=self.cluster.cluster_name, service_name=self.service_name)
        @property  # noqa
        def is_blue(self) -> bool:
            return AwsEcs._is_blue(self.service_name)
        @property  # noqa
        def is_green(self) -> bool:
            return AwsEcs._is_green(self.service_name)
        @property  # noqa
        def blue_or_green(self) -> str:
            return AwsEcs._blue_or_green(self.service_name)
        @property  # noqa
        def type(self) -> str:
            return AwsEcs._type(self.service_name)
        @property  # noqa
        def is_mirrored(self) -> bool:
            return (self.is_blue and self.task_definition.is_green) or (self.is_green and self.task_definition.is_blue)
        def get_annotation(self, dns: bool = False) -> str:  # noqa
            annotation = ""
            if self.type:
                if annotation:
                    annotation += " | "
                annotation = self.type.upper()
            env = None
            dns_cname = self.dns_cname if dns and AwsEcs._is_portal(self.service_name) else None
            if dns_cname:
                if dns_cname.startswith("data."):
                    env = "DATA"
                elif dns_cname.startswith("staging."):
                    env = "STAGING"
            if env:
                if annotation:
                    annotation += " | "
                annotation += env
            if self.blue_or_green:
                if annotation:
                    annotation += " | "
                annotation += self._ecs._terminal_color(self.blue_or_green.upper(),
                                                        self.blue_or_green, underline=False)
            if self.is_mirrored and False:
                if annotation:
                    annotation += " | "  # " ▶ "
                annotation += "MIRROR"
            return f"{annotation}" if annotation else ""
        @property  # noqa
        def dns_aname(self) -> Optional[str]:
            # Get ANAME/CNAME so we can definitively see which service is associated with data/staging.
            # From service (e.g. c4-ecs-blue-green-smaht-production-stack-SmahtgreenPortalService-aYgRm5cTsSu0),
            # get application load-balancer (EC2) target-group name (e.g. TargetGroupApplicationGreen), and then
            # the load-balancer name (e.g. smaht-productiongreen - though should be able to go directly to LB?),
            # and there get, e.g.: DNS: smaht-productiongreen-1114221794.us-east-1.elb.amazonaws.com (A Record).
            if self._dns_aname:
                return self._dns_aname
            try:
                if not (services := self._ecs._describe_services(self.cluster.cluster_name, self.service_name)):
                    return None
                if not (load_balancers := services[0].get("loadBalancers")):
                    return None
                boto_elb = BotoClient("elbv2")
                # For some reason have to get target-group name first and from there get the load-balancer info.
                for load_balancer in load_balancers:
                    if target_group_arn := load_balancer.get("targetGroupArn"):
                        if ((target_groups := boto_elb.describe_target_groups(TargetGroupArns=[target_group_arn])) and
                            (target_group := target_groups.get("TargetGroups")[0])):  # noqa
                            self._target_group_arn = target_group.get("TargetGroupArn")
                            if load_balancers := target_group.get("LoadBalancerArns"):
                                if load_balancer := boto_elb.describe_load_balancers(
                                        LoadBalancerArns=load_balancers)["LoadBalancers"]:
                                    load_balancer = load_balancer[0]
                                    self._dns_aname = load_balancer.get("DNSName")
                                    load_balancer_arn = load_balancer.get("LoadBalancerArn")
                                    listeners = boto_elb.describe_listeners(
                                        LoadBalancerArn=load_balancer_arn)["Listeners"]
                                    for listener in listeners:
                                        if listener.get("Port") == 443:
                                            ssl_certificate_arn = listener["Certificates"][0]["CertificateArn"]
                                            boto_acm = BotoClient("acm")
                                            certificate_info = boto_acm.describe_certificate(
                                                CertificateArn=ssl_certificate_arn)
                                            certificate = certificate_info.get("Certificate")
                                            self._certificate_expiration_date = certificate.get("NotAfter")
                                            if certificate_names := certificate.get("SubjectAlternativeNames"):
                                                self._dns_cname = certificate_names[0]
                                            break
                                    return self._dns_aname
            except Exception:
                return None
        @property  # noqa
        def dns_cname(self) -> Optional[str]:
            if self._dns_cname:
                return self._dns_cname
            if not self.dns_aname:
                return None
            return self._dns_cname
        @property  # noqa
        def global_env_bucket(self) -> Optional[str]:
            if not self._global_env_bucket:
                _ = self.dns_aname
            return self._global_env_bucket
        @property  # noqa
        def environment_name(self) -> Optional[str]:
            if not self._environment_name:
                _ = self.dns_aname
            return self._environment_name
        @property  # noqa
        def target_group_arn(self) -> Optional[str]:
            if not self._target_group_arn:
                _ = self.dns_aname
            return self._target_group_arn
        @property  # noqa
        def certificate_expiration_date(self) -> Optional[str]:
            if not self._certificate_expiration_date:
                _ = self.dns_aname
            return self._certificate_expiration_date
        def __str__(self) -> str:  # noqa
            annotation = self.get_annotation()
            return f"{self.service_name}{f' {annotation}' if annotation else ''}"

    class TaskDefinition:
        def __init__(self, task_definition_arn: str, ecs: Optional[AwsEcs] = None) -> None:
            self.task_definition_arn = task_definition_arn or ""
            self.task_definition_name = AwsEcs._nonarn_name(task_definition_arn)
            self.task_definition_name_unversioned = AwsEcs._unversioned_name(self.task_definition_name)
            self._ecs = ecs if isinstance(ecs, AwsEcs) else AwsEcs()
            self._ecs._note_name(self.task_definition_name)
        @property  # noqa
        def is_blue(self) -> bool:
            return AwsEcs._is_blue(self.task_definition_arn)
        @property  # noqa
        def is_green(self) -> bool:
            return AwsEcs._is_green(self.task_definition_arn)
        @property  # noqa
        def blue_or_green(self) -> str:
            return AwsEcs._blue_or_green(self.task_definition_arn)
        @property  # noqa
        def is_mirror(self) -> bool:
            # Fourfront only; for task defintion names like this:
            # c4-ecs-fourfront-production-stack-FourfrontbluePortal-Mirror-puoq5pF2shUR
            return self.blue_or_green and ("-Mirror-" in self.task_definition_name)
        @property  # noqa
        def type(self) -> str:
            return AwsEcs._type(self.task_definition_name)

        def get_container(self, service: AwsEcs.Service, noimage: bool = False, nogit: bool = False) -> dict:

            try:
                containers = self._ecs._boto_ecs.describe_task_definition(
                    taskDefinition=self.task_definition_name)["taskDefinition"].get("containerDefinitions")
            except Exception:
                return {}
            result = []

            # Mistake before in not getting correct image build info based on digest associated with running tasks.
            service_running_tasks_image_digest = None
            if service_running_tasks := service.running_tasks:
                if service_running_tasks := service._ecs._boto_ecs.describe_tasks(cluster=service.cluster.cluster_name,
                                                                                  tasks=service_running_tasks)["tasks"]:
                    for service_running_task in service_running_tasks:
                        for service_running_task_container in service_running_task.get("containers", []):
                            if digest := service_running_task_container.get("imageDigest"):
                                if service_running_tasks_image_digest is None:
                                    service_running_tasks_image_digest = digest
                                elif service_running_tasks_image_digest != digest:
                                    print(f"WARNING: Different running task image digest"
                                          f" values for service: {service.service_name}")
                                    service_running_tasks_image_digest = None
                                    break

            for container in containers:
                container_name = container.get("name")
                container_identity = None
                elasticsearch_server = None ; redis_server = None  # noqa
                image = None ; image_repo = None ; image_tag = None  # noqa
                image_size = None ; image_pushed_at = None ; image_digest = None  # noqa
                git_repo = None ; git_branch = None ; git_commit = None  # noqa
                git_commit_date = None ; git_commit_latest = None ; git_commit_latest_date = None  # noqa
                build_project = None  # noqa
                if envs := container.get("environment"):
                    # Get the IDENTITY associated with thie task-definition/container.
                    for env in envs:
                        if env.get("name") == "IDENTITY":
                            if container_identity := env.get("value"):
                                elasticsearch_server = self._ecs._get_elasticsearch_server(container_identity)
                                redis_server = self._ecs._get_redis_server(container_identity)
                                database_server = self._ecs._get_database_server(container_identity)
                                global_env_bucket = self._ecs._get_global_env_bucket(container_identity)
                                environment_name = self._ecs._get_environment_name(container_identity)
                                s3_encrypt_key = self._ecs._get_s3_encrypt_key(container_identity)
                                s3_encrypt_key_id = self._ecs._get_s3_encrypt_key_id(container_identity)
                                auth0_client = self._ecs._get_auth0_client(container_identity)
                                auth0_secret = self._ecs._get_auth0_secret(container_identity)
                                auth0_domain = self._ecs._get_auth0_domain(container_identity)
                            break
                # Get the ECR image associated with this container; looks like this:
                # 643366669028.dkr.ecr.us-east-1.amazonaws.com/fourfront-production
                if (not noimage) and (image := container.get("image")):
                    image_info = self._ecs._get_ecr_image_info(image)
                    image = image_info["name"]
                    image_repo = image_info["repo"]
                    image_tag = image_info["tag"]
                    image_size = image_info["size"]
                    image_pushed_at = format_datetime(image_info["pushed"], notz=True)
                    image_digest = service_running_tasks_image_digest or image_info["digest"]
                    if ((not nogit) and
                        (build_info := get_image_build_info(image_repo, image_tag, image_digest))):  # noqa
                        build_project = build_info.get("build_project")
                        git_repo = build_info.get("repo")
                        git_branch = build_info.get("branch")
                        git_commit = build_info.get("commit")
                        git_commit_date = build_info.get("commit_date")
                        git_commit_latest = build_info.get("commit_latest")
                        git_commit_latest_date = build_info.get("commit_latest_date")
                result.append({"name": container_name,
                               "identity": container_identity,
                               "build_project": build_project,
                               "image": image,
                               "image_repo": image_repo,
                               "image_tag": image_tag,
                               "image_size": image_size,
                               "image_pushed_at": image_pushed_at,
                               "image_digest": image_digest,
                               "git_repo": git_repo,
                               "git_commit": git_commit,
                               "git_commit_date": git_commit_date,
                               "git_commit_latest": git_commit_latest,
                               "git_commit_latest_date": git_commit_latest_date,
                               "git_branch": git_branch,
                               "elasticsearch": elasticsearch_server,
                               "database": database_server,
                               "redis": redis_server,
                               "global_env_bucket": global_env_bucket,
                               "environment_name": environment_name,
                               "s3_encrypt_key": s3_encrypt_key,
                               "s3_encrypt_key_id": s3_encrypt_key_id,
                               "auth0_client": auth0_client,
                               "auth0_secret": auth0_secret,
                               "auth0_domain": auth0_domain})

            # We just return one - that is all we normally have. TODO: Ask if this is for us alway true.
            return result[0] if result else {}

        @property  # noqa
        def annotation(self) -> str:
            annotation = ""
            if self.type:
                if annotation:
                    annotation += " | "
                annotation = self.type.upper()
            if self.blue_or_green:
                if annotation:
                    annotation += " | "
                annotation += self._ecs._terminal_color(self.blue_or_green.upper(), self.blue_or_green)
            return f"{annotation}" if annotation else ""
        def __str__(self) -> str:  # noqa
            annotation = self.annotation
            return f"{self.task_definition_name}{f' {annotation}' if annotation else ''}"

    class TaskDefinitionSwap:
        def __init__(self, service: AwsEcs.Service, new_task_definition: AwsEcs.TaskDefinition) -> None:
            self.service = service
            self.new_task_definition = new_task_definition

    def __init__(self, blue_green: Optional[Union[Literal[AwsEcs.BLUE_OR_GREEN], bool]] = False,
                 nocolor: bool = False, boto_ecs: Optional[BotoEcs] = None) -> None:
        self._boto_ecs = BotoClient("ecs") if not isinstance(boto_ecs, BotoEcs) else boto_ecs
        self._blue_green = blue_green
        self._clusters = None
        self._task_definitions = None
        self._nocolor = nocolor is True
        self._names = []

    @property
    def clusters(self) -> List[Cluster]:
        if self._clusters is None:
            self._clusters = []
            for cluster_arn in self._list_clusters():
                if (((self._blue_green == AwsEcs.BLUE) and (not AwsEcs._is_blue(cluster_arn))) or
                    ((self._blue_green == AwsEcs.GREEN) and (not AwsEcs._is_green(cluster_arn))) or
                    ((self._blue_green is True) and (not AwsEcs._blue_or_green(cluster_arn)))):  # noqa
                    continue
                self._clusters.append(AwsEcs.Cluster(cluster_arn, ecs=self))
            self._clusters = sorted(self._clusters, key=lambda item: (not item.blue_or_green, item.cluster_name))
        return self._clusters

    def find_task_definition(self, task_definition: Union[str, TaskDefinition]) -> Optional[AwsEcs.TaskDefinition]:
        if isinstance(task_definition, AwsEcs.TaskDefinition):
            task_definition_name = self._unversioned_name(task_definition.task_definition_name)
        elif isinstance(task_definition, str):
            task_definition_name = self._unversioned_name(self._nonarn_name(task_definition))
        else:
            return None
        for cluster in self.clusters:
            for service in cluster.services:
                if (self._unversioned_name(service.task_definition.task_definition_name) == task_definition_name):
                    return service.task_definition
        return None

    @property
    def task_definitions(self) -> List[TaskDefinition]:
        if self._task_definitions is None:
            try:
                task_definitions = []
                if task_definition_arns := self._boto_ecs.list_task_definitions().get("taskDefinitionArns"):
                    for task_definition_arn in task_definition_arns:
                        task_definitions.append(AwsEcs.TaskDefinition(task_definition_arn))
            except Exception:
                pass
            task_definitions_latest = []
            last_task_definition = None
            for task_definition in sorted(task_definitions, key=lambda item: item.task_definition_arn, reverse=True):
                if last_task_definition and (task_definition.task_definition_name_unversioned ==
                                             last_task_definition.task_definition_name_unversioned):
                    continue
                task_definitions_latest.append(task_definition)
                last_task_definition = task_definition
            self._task_definitions = task_definitions_latest
        return self._task_definitions

    @property
    def unassociated_task_definition_names(self) -> List[str]:
        try:
            unassociated_task_definition_names = []
            if task_definition_arns := self._boto_ecs.list_task_definitions().get("taskDefinitionArns"):
                for task_definition_arn in task_definition_arns:
                    task_definition_name = self._unversioned_name(self._nonarn_name(task_definition_arn))
                    if self.find_task_definition(task_definition_name) is None:
                        if task_definition_name not in unassociated_task_definition_names:
                            unassociated_task_definition_names.append(task_definition_name)
        except Exception:
            pass
        return sorted(unassociated_task_definition_names)

    def format_name(self, value: str, versioned: bool = True, shortened: bool = False) -> str:
        if versioned is False:
            value = self._unversioned_name(value)
        if (shortened is True) and isinstance(value, str) and (prefix := AwsEcs._longest_common_prefix(self._names)):
            if value.startswith(prefix):
                value = value[len(prefix):]
        return value

    @property
    def account(self) -> Optional[object]:
        boto_sts = BotoClient("sts")
        boto_iam = BotoClient("iam")
        try:
            account_number = boto_sts.get_caller_identity()["Account"]
            account_alias = None
            try:
                account_alias = boto_iam.list_account_aliases()["AccountAliases"][0]
            except Exception:
                pass
            return namedtuple("aws_account_info", ["account_number", "account_alias"])(account_number, account_alias)
        except Exception:
            return None

    def identity_swap_plan(self) -> Tuple[Optional[List[AwsEcs.TaskDefinitionSwap]], Optional[str]]:
        swaps, error = self._identity_swap(swap=False)
        if error:
            return None, error
        return swaps, None

    def identity_swap(self) -> Tuple[Optional[AwsEcs], Optional[str]]:
        identity_swapped, error = self._identity_swap(swap=True)
        if error:
            return None, error
        return identity_swapped, None

    def _identity_swap(self, swap: bool = False) -> Tuple[Optional[Union[List[AwsEcs.TaskDefinitionSwap],
                                                                         AwsEcs]], Optional[str]]:

        # Sanity check that all clusters are in the same normal/mirrored state.
        last_cluster = None
        for cluster in self.clusters:
            if (last_cluster is not None) and (last_cluster.is_mirrored != cluster.is_mirrored):
                return None, "Not all clusters are in the same normal/mirrored state; something is wrong."
            last_cluster = cluster

        if self._is_fourfront:
            return self._identity_swap_fourfront(swap)

        # Get the blue cluster.
        if not (blue_cluster := [cluster for cluster in self.clusters if cluster.is_blue]):
            return None, "Blue cluster not found."
        elif len(blue_cluster) > 1:
            return None, f"Mutliple ({len(blue_cluster)}) blue clusters not found: TODO"
        blue_cluster = blue_cluster[0]
        blue_services = blue_cluster.services

        # Get the green cluster.
        if not (green_cluster := [cluster for cluster in self.clusters if cluster.is_green]):
            return None, "Green cluster not found."
        elif len(green_cluster) > 1:
            return None, f"Mutliple ({len(green_cluster)}) green clusters not found: TODO"
        green_cluster = green_cluster[0]
        green_services = green_cluster.services

        # Sanity check service count.
        if len(blue_services) != len(green_services):
            return None, f"Different number of blue ({len(blue_services)}) and green ({len(green_services)}) services."

        # Sanity check that both blue/green cluster are in the same mirrored or normal state.
        if blue_cluster.is_mirrored != green_cluster.is_mirrored:
            return None, "Blue and green clusters are in an inconsistent mirrored state."

        if swap is not True:
            swaps = []

        for service_type in AwsEcs.TYPES:
            blue_services_of_type = [service for service in blue_services if service.type == service_type]
            green_services_of_type = [service for service in green_services if service.type == service_type]
            if len(blue_services_of_type) != len(green_services_of_type):
                return None, (f"Different number of blue ({len(blue_services)}) and"
                              f" green ({len(green_services)}) {service_type.upper()} services.")
            for index, blue_service in enumerate(blue_services_of_type):
                green_service = green_services_of_type[index]
                blue_service_task_definition = blue_service.task_definition
                green_service_task_definition = green_service.task_definition
                if swap is True:
                    # This is the actual swap (of the data - not actually in AWS of course) right here:
                    green_service.task_definition = blue_service_task_definition
                    blue_service.task_definition = green_service_task_definition
                else:
                    # Record the proposed swap in list of TaskDefinitionSwap objects.
                    swaps.append(AwsEcs.TaskDefinitionSwap(green_service, blue_service_task_definition))
                    swaps.append(AwsEcs.TaskDefinitionSwap(blue_service, green_service_task_definition))

        if swap is True:
            return self, None
        else:
            return swaps, None

    def _identity_swap_fourfront(self, swap: bool = False) -> Tuple[Optional[Union[List[AwsEcs.TaskDefinitionSwap],
                                                                                   AwsEcs]], Optional[str]]:
        # Get the blue cluster.
        if not (blue_cluster := [cluster for cluster in self.clusters if cluster.is_blue]):
            return None, "Blue cluster not found."
        elif len(blue_cluster) > 1:
            return None, f"Mutliple ({len(blue_cluster)}) blue clusters not found: TODO"
        blue_cluster = blue_cluster[0]
        blue_services = blue_cluster.services

        # Get the green cluster.
        if not (green_cluster := [cluster for cluster in self.clusters if cluster.is_green]):
            return None, "Green cluster not found."
        elif len(green_cluster) > 1:
            return None, f"Mutliple ({len(green_cluster)}) green clusters not found: TODO"
        green_cluster = green_cluster[0]
        green_services = green_cluster.services

        # Sanity check service count.
        if len(blue_services) != len(green_services):
            return None, f"Different number of blue ({len(blue_services)}) and green ({len(green_services)}) services."

        # Sanity check that both blue/green cluster are in the same mirrored or normal state.
        if (is_mirrored := blue_cluster.is_mirrored) != green_cluster.is_mirrored:
            return None, "Blue and green clusters are in an inconsistent mirrored state."

        task_definitions = self.task_definitions

        if swap is not True:
            swaps = []

        # When we are in a mirrored state; for example, for a service named like this:
        # - c4-ecs-fourfront-production-stack-FourfrontgreenPortalService-3fSt8wKOVlzF
        # we have service tasks named like this:
        # - c4-ecs-fourfront-production-stack-FourfrontbluePortal-Mirror-puoq5pF2shUR
        # so for the identity-swap we will repoint the service tasks named like this:
        # - c4-ecs-fourfront-production-stack-FourfrontgreenPortal-72ALIPiA0sNC
        # so that we will then end up in a normal state.

        # When we are in a normal state; for example, for a service named like this:
        # - c4-ecs-fourfront-production-stack-FourfrontgreenPortalService-3fSt8wKOVlzF
        # we have service tasks named like this:
        # - c4-ecs-fourfront-production-stack-FourfrontgreenPortal-72ALIPiA0sNC
        # so for the identity-swap we will repoint the service tasks named like this:
        # - c4-ecs-fourfront-production-stack-FourfrontbluePortal-Mirror-puoq5pF2shUR
        # so that we will then end up in a mirrored state.

        for cluster in self.clusters:
            for service in cluster.services:
                for task_definition in task_definitions:
                    if ((task_definition.type == service.task_definition.type) and
                        (task_definition.blue_or_green != service.task_definition.blue_or_green) and
                        (task_definition.is_mirror == (not is_mirrored))):  # noqa
                        if swap is True:
                            # This is the actual swap (of the data here - not actually in AWS of course) right here:
                            service.task_definition = task_definition
                        else:
                            # Record the proposed swap in list of TaskDefinitionSwap objects.
                            swaps.append(AwsEcs.TaskDefinitionSwap(service, task_definition))
                        break

        if swap is True:
            return self, None
        else:
            return swaps, None

    def print(self, shortened_names: bool = False, versioned_names: bool = False, nodns: bool = False,
              nocontainer: bool = False, noimage: bool = False, nogit: bool = False, notasks: bool = False,
              nohealth: bool = False, nouptime: bool = False, noasync: bool = False,
              show: bool = False, verbose: bool = False) -> AwsEcs:
        cluster_results = {}
        def print_cluster(cluster: AwsEcs.Cluster, noprint: bool = False):  # noqa
            nonlocal cluster_results, shortened_names, versioned_names, nodns
            nonlocal nocontainer, noimage, nogit, notasks, nohealth, nouptime, show, verbose
            cluster_lines = cluster.print_cluster(
                shortened_names=shortened_names, versioned_names=versioned_names, nodns=nodns,
                nocontainer=nocontainer, noimage=noimage, nogit=nogit, notasks=notasks,
                nohealth=nohealth, nouptime=nouptime, show=show, verbose=verbose, noprint=noprint)
            cluster_results[cluster] = cluster_lines
        if noasync is not True:
            functions = [lambda cluster=cluster: print_cluster(cluster, noprint=True) for cluster in self.clusters]
            run_concurrently(functions, nthreads=8)
            cluster_results = dict(sorted(cluster_results.items(),
                                          key=lambda cluster: (not cluster[0].blue_or_green, cluster[0].cluster_name)))
            for cluster in cluster_results:
                cluster_lines = cluster_results[cluster]
                for cluster_line in cluster_lines:
                    print(cluster_line)
        else:
            for cluster in self.clusters:
                print_cluster(cluster)
        print("")

    def _list_clusters(self) -> List[str]:
        try:
            return self._boto_ecs.list_clusters().get("clusterArns", [])
        except Exception:
            return []

    def _list_services(self, cluster_name: str) -> List[dict]:
        try:
            return self._boto_ecs.list_services(cluster=cluster_name).get("serviceArns", [])
        except Exception:
            return []

    def _list_running_tasks(self, cluster_name: str, service_name: Optional[str] = None) -> List[dict]:
        try:
            if isinstance(service_name, str) and service_name:
                return self._boto_ecs.list_tasks(cluster=cluster_name,
                                                 serviceName=service_name, desiredStatus="RUNNING")["taskArns"]
            else:
                return self._boto_ecs.list_tasks(cluster=cluster_name, desiredStatus="RUNNING")["taskArns"]
        except Exception:
            return []

    def _describe_services(self, cluster_name: str, service_names: List[str]) -> List[str]:
        try:
            if isinstance(service_names, str):
                service_names = [service_names]
            return self._boto_ecs.describe_services(cluster=cluster_name, services=service_names)["services"]
        except Exception:
            return []

    @lru_cache
    def _get_elasticsearch_server(self, identity: str) -> Optional[str]:
        return self._get_identity_secrets(identity).get("ENCODED_ES_SERVER")

    @lru_cache
    def _get_database_server(self, identity: str) -> Optional[str]:
        return self._get_identity_secrets(identity).get("RDS_HOSTNAME")

    @lru_cache
    def _get_redis_server(self, identity: str) -> Optional[str]:
        if redis := self._get_identity_secrets(identity).get("ENCODED_REDIS_SERVER"):
            return self._unversioned_name(redis.replace("rediss://", ""))
        return None

    @lru_cache
    def _get_global_env_bucket(self, identity: str) -> Optional[str]:
        return self._get_identity_secrets(identity).get("GLOBAL_ENV_BUCKET")

    @lru_cache
    def _get_environment_name(self, identity: str) -> Optional[str]:
        return (self._get_identity_secrets(identity).get("ENV_NAME") or
                self._get_identity_secrets(identity).get("ENCODED_BS_ENV"))

    @lru_cache
    def _get_s3_encrypt_key(self, identity: str) -> Optional[str]:
        return self._get_identity_secrets(identity).get("S3_ENCRYPT_KEY")

    @lru_cache
    def _get_s3_encrypt_key_id(self, identity: str) -> Optional[str]:
        return self._get_identity_secrets(identity).get("ENCODED_S3_ENCRYPT_KEY_ID")

    @lru_cache
    def _get_auth0_client(self, identity: str) -> Optional[str]:
        return self._get_identity_secrets(identity).get("ENCODED_AUTH0_CLIENT")

    @lru_cache
    def _get_auth0_secret(self, identity: str) -> Optional[str]:
        return self._get_identity_secrets(identity).get("ENCODED_AUTH0_SECRET")

    @lru_cache
    def _get_auth0_domain(self, identity: str) -> Optional[str]:
        return self._get_identity_secrets(identity).get("ENCODED_AUTH0_DOMAIN")

    @lru_cache
    def _get_identity_secrets(self, identity: str) -> dict:
        try:
            return get_identity_secrets(identity_name=identity)
        except Exception:
            return {}

    @lru_cache
    def _get_ecr_image_info(self, image_name: str) -> dict:
        try:
            account_id, repo_with_tag = image_name.split(".")[0], image_name.split("/")[-1]
            repo_name, image_tag = repo_with_tag.split(":")
            ecr = BotoClient("ecr")
            response = ecr.describe_images(repositoryName=repo_name,
                                           imageIds=[{"imageTag": image_tag}], registryId=account_id)["imageDetails"][0]
            return {
                "name": os.path.basename(image_name),
                "repo": repo_name,
                "tag": image_tag,
                "size": response["imageSizeInBytes"],
                "pushed": response["imagePushedAt"],
                "digest": response["imageDigest"]
            }
        except Exception:
            return {}

    @staticmethod
    def _is_blue(value: str) -> bool:
        return AwsEcs._blue_or_green(value) == AwsEcs.BLUE

    @staticmethod
    def _is_green(value: str) -> bool:
        return AwsEcs._blue_or_green(value) == AwsEcs.GREEN

    @staticmethod
    def _blue_or_green(value: str) -> Optional[Literal[AwsEcs.BLUE_OR_GREEN]]:
        if isinstance(value, str) and (value := value.lower()):
            blues = value.count(AwsEcs.BLUE)
            greens = value.count(AwsEcs.GREEN)
            if (blues > 0) or (greens > 0):
                return AwsEcs.BLUE if blues > greens else (AwsEcs.GREEN if greens > blues else None)
            return None

    @staticmethod
    def _is_portal(value: str) -> bool:
        return AwsEcs._type(value) == AwsEcs.PORTAL

    @staticmethod
    def _is_indexer(value: str) -> bool:
        return AwsEcs._type(value) == AwsEcs.INDEXER

    @staticmethod
    def _is_ingester(value: str) -> bool:
        return AwsEcs._type(value) == AwsEcs.INGESTER

    @property
    def _is_fourfront(self) -> bool:
        for cluster in self.clusters:
            if "fourfront" not in cluster.cluster_name.lower():
                return False
        return True

    @staticmethod
    def _type(value: str) -> Optional[Literal[AwsEcs.TYPES]]:
        if isinstance(value, str):
            if AwsEcs.PORTAL.lower() in value.lower():
                return AwsEcs.PORTAL
            elif AwsEcs.INDEXER.lower() in value.lower():
                return AwsEcs.INDEXER
            elif AwsEcs.INGESTER.lower() in value.lower():
                return AwsEcs.INGESTER
        return None

    @staticmethod
    def _nonarn_name(value: str) -> str:
        return value.split("/")[-1] if isinstance(value, str) and "/" in value else value

    @staticmethod
    def _unversioned_name(value: str) -> str:
        if isinstance(value, str) and ((colon := value.rfind(":")) > 0):
            return value[:colon]
        return value

    def _note_name(self, value: str) -> None:
        if isinstance(value, str) and value and (value.lower() != "default"):
            self._names.append(value)

    @staticmethod
    def _longest_common_prefix(strings: List[str]) -> str:
        if not (isinstance(strings, list) and strings and isinstance(prefix := strings[0], str)):
            return ""
        for value in strings[1:]:
            if not isinstance(value, str):
                return ""
            while not value.startswith(prefix):
                prefix = prefix[:-1]
                if not prefix:
                    return ""
        return prefix

    @staticmethod
    def _get_blue_or_green_from_health(health: Optional[dict]) -> Optional[Literal[AwsEcs.BLUE_OR_GREEN]]:
        if isinstance(health, dict):
            return AwsEcs._blue_or_green(health.get("beanstalk_env"))
        return None

    @staticmethod
    def _get_health_page(url: str) -> Optional[dict]:
        if isinstance(url, str):
            if not (url := url.lower()).startswith("http"):
                url = f"https://{url}"
            url = f"{url}/health?format=json"
            try:
                return requests.get(url).json()
            except Exception:
                if url.startswith("https://"):
                    url = url.replace("https://", "http://")
                else:
                    url = url.replace("http://", "https://")
                try:
                    return requests.get(url).json()
                except Exception:
                    return None

    def _terminal_color(self, value: str, color: str,
                        bold: bool = True, underline: bool = True, dark: bool = False) -> str:
        if self._nocolor:
            return value
        attributes = []
        if bold:
            attributes.append("bold")
        if underline:
            attributes.append("underline")
        if dark:
            attributes.append("dark")
        return colored(value, color.lower(), attrs=attributes)


class AwsS3:

    def __init__(self, bucket: str) -> None:
        self._bucket = bucket if isinstance(bucket, str) and (bucket := bucket.strip()) else None
        self._boto_s3 = BotoClient("s3") if self._bucket else None

    def load_json_file(self, key: str) -> Optional[dict]:
        try:
            stream = io.BytesIO()
            self._boto_s3.download_fileobj(Fileobj=stream, Bucket=self._bucket, Key=key)
            return json.loads(stream.getvalue())
        except Exception:
            return {}


def usage() -> None:
    print("usage: awsecs [--bluegreen] [--swap] [--short] [--versioned] [--aws aws-profile-name]")
    exit(1)


def main():

    blue_green = False
    shortened_names = False
    versioned_names = False
    identity_swap = False
    show_unassociated_task_definitions = False
    nodns = False
    nocontainer = False
    noimage = False
    nogit = False
    notasks = False
    nohealth = False
    nouptime = False
    nocolor = False
    noasync = False
    show = False
    verbose = False

    argi = 0 ; argn = len(argv := sys.argv[1:])  # noqa
    while argi < argn:
        arg = argv[argi] ; argi += 1  # noqa
        if arg in ["--bluegreen", "-bluegreen", "bluegreen", "--greenblue",
                   "-greenblue", "greenblue", "--bg", "-bg", "bg", "--gb", "-gb", "gb"]:
            # Show only blue/green clusters/services/task-definitions.
            blue_green = True
        elif arg in ["--blue", "-blue", "blue", "--blue"]:
            # Show only blue clusters/services/task-definitions.
            blue_green = AwsEcs.BLUE
        elif arg in ["--green", "-green", "green", "--green"]:
            # Show only green clusters/services/task-definitions.
            blue_green = AwsEcs.GREEN
        elif arg in ["--short", "-short", "short"]:
            # Display shortened names for easier viewing if possible; removes longest common prefix.
            shortened_names = True
        elif arg in ["--versioned", "-versioned", "versioned"]:
            # Dot not lop off the ":n" from the end of task definition names.
            versioned_names = True
        elif arg in ["--identity-swap", "-identity-swap", "identity-swap", "--swap", "-swap", "swap"]:
            # Show identity swap plan - not able to actually do anything.
            identity_swap = True
        elif arg in ["--nodns", "-nodns", "nodns"]:
            nodns = True
            nohealth = True
        elif arg in ["--nocontainer", "-nocontainer", "nocontainer"]:
            nocontainer = True
        elif arg in ["--quick", "-quick", "quick", "--q", "-q"]:
            nocontainer = True
            nodns = True
        elif arg in ["--noimage", "-noimage", "noimage"]:
            noimage = True
        elif arg in ["--nogit", "-nogit", "nogit"]:
            nogit = True
        elif arg in ["--notasks", "-notasks", "--notask", "-notask"]:
            notasks = True
        elif arg in ["--nohealth", "-nohealth"]:
            nohealth = True
        elif arg in ["--nocolor", "-nocolor", "nocolor"]:
            nocolor = True
        elif arg in ["--nouptime", "-nouptime", "nouptime"]:
            nouptime = True
        elif arg in ["--show", "-show", "show"]:
            show = True
        elif arg in ["--unassociated", "-unassociated", "--unassoc", "-unassoc"]:
            show_unassociated_task_definitions = True
        elif arg in ["--noasync", "-noasync", "noasync"]:
            noasync = True
        elif arg in ["--github-token", "-github-token", "--git", "-git"]:
            if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                usage()
            os.environ["GITHUB_TOKEN"] = arg
            argi += 1
        elif arg in ["--verbose", "-verbose", "--verbose", "-verbose", "--v", "-v"]:
            verbose = True
        elif arg in ["--aws", "-aws", "--env", "-env"]:
            # Profile name from ~/.aws/config file.
            if (argi >= argn) or (aws_profile := argv[argi]).startswith("-"):
                usage()
            os.environ["AWS_PROFILE"] = aws_profile
            argi += 1
        elif arg in ["--version", "-version"]:
            print(f"hms-utils version: {get_version()}")
            exit(0)
        else:
            usage()

    if identity_swap:
        blue_green = True

    ecs = AwsEcs(blue_green=blue_green, nocolor=nocolor)

    if not (ecs_account := ecs.account):
        print("AWS credentials do not appear to be working.")
        exit(1)

    print(f"Showing current ECS"
          f"{' blue/green' if (blue_green is True) else (f' {blue_green}' if blue_green else '')} cluster info"
          f" for AWS account: {ecs_account.account_number}"
          f"{f' ({ecs_account.account_alias})' if ecs_account.account_alias else ''} ...")

    ecs.print(shortened_names=shortened_names, versioned_names=versioned_names,
              nodns=nodns, nocontainer=nocontainer, noimage=noimage, nogit=nogit,
              notasks=notasks, nohealth=nohealth, nouptime=nouptime, show=show, noasync=noasync, verbose=verbose)

    if identity_swap:
        swaps, error = ecs.identity_swap_plan()
        if error:
            print(error)
            exit(1)
        if ecs.clusters[0].is_mirrored:
            source_state = "MIRRORED"
            target_state = "NORMAL"
        else:
            source_state = "NORMAL"
            target_state = "MIRRORED"
        print(f"Showing proposed ECS identity swap plan for AWS account: {ecs_account.account_number}"
              f"{f' ({ecs_account.account_alias})' if ecs_account.account_alias else ''}"
              f" | {source_state} {chars.rarrow} {target_state}")
        for swap in swaps:
            service_name = ecs.format_name(swap.service.service_name,
                                           versioned=versioned_names, shortened=shortened_names)
            service_annotation = swap.service.get_annotation()
            task_definition_name = ecs.format_name(swap.service.task_definition.task_definition_name,
                                                   versioned=True, shortened=shortened_names)
            task_definition_annotation = swap.service.task_definition.annotation
            new_task_definition_name = ecs.format_name(swap.new_task_definition.task_definition_name,
                                                       versioned=True, shortened=shortened_names)
            new_task_definition_annotation = swap.new_task_definition.annotation
            print(f"\n- SERVICE: {service_name}{f' | {service_annotation}' if service_annotation else ''}")
            print(f"  - CURRENT TASK: {task_definition_name}"
                  f"{f' | {task_definition_annotation}' if task_definition_annotation else ''}")
            print(f"     ▶▶ NEW TASK: {new_task_definition_name}"
                  f"{f' | {new_task_definition_annotation}' if new_task_definition_annotation else ''}")
        print()
        print(f"It would look like this after the swap: {target_state}")
        ecs_swapped, error = ecs.identity_swap()
        if error:
            print(error)
            exit(1)
        # TODO: Something going wrong when showing image info in swap state - wires crossed; disable for
        # now; might be better like this anyways - too much noise; don't want all that to see swap state.
        ecs_swapped.print(shortened_names=shortened_names, versioned_names=versioned_names,
                          nodns=nodns, nocontainer=True, noimage=noimage, nogit=nogit,
                          notasks=notasks, nohealth=True, nouptime=True, show=show,
                          noasync=noasync, verbose=verbose)

    if show_unassociated_task_definitions:
        if unassociated_task_definition_names := ecs.unassociated_task_definition_names:
            print("Task definitions unassociated with any service:\n")
            for unassociated_task_definition_name in unassociated_task_definition_names:
                print(f"- {unassociated_task_definition_name}")


if __name__ == "__main__":
    main()



import calendar
from datetime import date, datetime
from dateutil.relativedelta import relativedelta
from typing import Any, Callable, List, Optional, Tuple, Union
from urllib.parse import urlencode
from dcicutils.datetime_utils import parse_datetime_string

AGGREGATION_FIELD_RELEASE_DATE = "file_status_tracking.released"
AGGREGATION_FIELD_CELL_LINE = "file_sets.libraries.analytes.samples.sample_sources.cell_line.code"
AGGREGATION_FIELD_DONOR = "donors.display_title"
AGGREGATION_FIELD_FILE_DESCRIPTOR = "release_tracker_description"

AGGREGATIONS = [
    AGGREGATION_FIELD_RELEASE_DATE,
    AGGREGATION_FIELD_CELL_LINE,
    AGGREGATION_FIELD_DONOR,
    AGGREGATION_FIELD_FILE_DESCRIPTOR
]

AGGREGATIONS_BY_CELL_LINE = [
    AGGREGATION_FIELD_RELEASE_DATE,
    AGGREGATION_FIELD_CELL_LINE,
    AGGREGATION_FIELD_FILE_DESCRIPTOR
]

AGGREGATIONS_BY_DONOR = [
    AGGREGATION_FIELD_RELEASE_DATE,
    AGGREGATION_FIELD_DONOR,
    AGGREGATION_FIELD_FILE_DESCRIPTOR
]

AGGREGATION_MAX_BUCKETS = 30
AGGREGATION_NO_VALUE = "No value"

DEFAULT_FILE_TYPE = "OutputFile"
DEFAULT_FILE_STATUS = "released"
DEFAULT_FILE_CATEGORY_IGNORED = "Quality Control"


def create_query_for_files(
        type: Optional[Union[str, List[str]]] = None,
        status: Optional[Union[str, List[str]]] = None,
        qcs: bool = False,
        from_date: Optional[str] = None,
        thru_date: Optional[str] = None,
        nmonths: Optional[str] = None,
        date_property_name: Optional[str] = None,
        aggregation: Optional[Union[str, List[str], dict]] = None,
        max_buckets: Optional[int] = None,
        missing_value: Optional[str] = None) -> Tuple[Optional[str], Optional[dict]]:

    global AGGREGATION_FIELD_RELEASE_DATE

    types = []
    if type is None:
        type = [DEFAULT_FILE_TYPE]
    if not isinstance(type, list):
        type = [type]
    for item in type:
        if isinstance(item, str) and (item := item.strip()) and (item not in types):
            types.append(item)

    statuses = []
    if status is None:
        status = [DEFAULT_FILE_STATUS]
    if not isinstance(status, list):
        status = [status]
    for item in status:
        if isinstance(item, str) and (item := item.strip()) and (item not in statuses):
            statuses.append(item)

    categories = []
    if qcs is not True:
        categories.append(f"!{DEFAULT_FILE_CATEGORY_IGNORED}")

    from_date, thru_date = parse_date_related_arguments(from_date, thru_date, nmonths=nmonths, strings=True)
    if from_date or thru_date:
        if not (isinstance(date_property_name, str) and (date_property_name := date_property_name.strip())):
            date_property_name = AGGREGATION_FIELD_RELEASE_DATE

    if not isinstance(aggregation, dict):
        aggregations = []
        if not isinstance(aggregation, list):
            aggregation = [aggregation]
        for item in aggregation:
            if isinstance(item, str) and (item := item.strip()) and (item not in aggregations):
                aggregations.append(item)
        def create_field_aggregation(field: str) -> Optional[dict]:  # noqa
            if field == AGGREGATION_FIELD_RELEASE_DATE:
                return {
                    "date_histogram": {
                        "field": f"embedded.{field}",
                        "calendar_interval": "month",
                        "format": "yyyy-MM",
                        "missing": "1970-01",
                        "order": {"_key": "desc"}
                    }
                }
        aggregation_definition = create_elasticsearch_aggregation_definition(
            aggregations,
            max_buckets=max_buckets,
            missing_value=missing_value,
            create_field_aggregation=create_field_aggregation)
    else:
        aggregation_definition = aggregation

    query_parameters = {
        "type": types if types else None,
        "status": status if status else None,
        "data_category": categories if categories else None,
        f"{date_property_name}.from": from_date if from_date else None,
        f"{date_property_name}.to": thru_date if from_date else None,
        "from": 0,
        "limit": 0,

    }
    query_parameters = {key: value for key, value in query_parameters.items() if value is not None}

    query_string = urlencode(query_parameters, True)
    # Hackishness to change "=!" to "!=" in search_param_lists value for e.g.: "data_category": ["!Quality Control"]
    query_string = query_string.replace("=%21", "%21=")

    return query_string, aggregation_definition


def create_elasticsearch_aggregation_definition(fields: List[str],
                                                max_buckets: Optional[int] = None,
                                                missing_value: Optional[str] = None,
                                                create_field_aggregation: Optional[Callable] = None) -> dict:
    global AGGREGATION_NO_VALUE

    if not (isinstance(fields, list) and fields and isinstance(field := fields[0], str) and field):
        return {}
    if not isinstance(missing_value, str):
        missing_value = AGGREGATION_NO_VALUE
    if not (isinstance(max_buckets, int) and (max_buckets > 0)):
        max_buckets = 30

    if not (callable(create_field_aggregation) and
            isinstance(field_aggregation := create_field_aggregation(field), dict)):
        field_aggregation = {
            "terms": {
                "field": f"embedded.{field}.raw",
                "missing": missing_value,
                "size": max_buckets
            }
        }

    aggregation = {field: field_aggregation}
    aggregation[field]["meta"] = {"field_name": field}

    if nested_aggregation := create_elasticsearch_aggregation_definition(
            fields[1:], max_buckets=max_buckets,
            missing_value=missing_value,
            create_field_aggregation=create_field_aggregation):
        aggregation[field]["aggs"] = nested_aggregation

    return aggregation


def parse_date_related_arguments(
        from_date: Optional[Union[str, datetime, date]],
        thru_date: Optional[Union[str, datetime, date]],
        nmonths: Optional[Union[str, int]] = None,
        strings: bool = False) -> Tuple[Optional[Union[str, datetime]], Optional[Union[str, datetime]]]:

    """
    Returns from/thru dates based on the given from/thru date arguments and optional nmonths argument.
    Given dates may be datetime or date objects or strings. Returned dates are datetime objects, or
    if the the given strings arguments is True, then strings (formatted as YYYY-MM-DD).

    If both of the given from/thru dates are specified/valid then those are returned
    and the given nmonths argument is not used.

    If only the given from date is specified then a None thru date is returned, UNLESS the given nmonths
    argument represents a positive integer, in which case the returned thru date will be nmonths months
    subsequent to the given from date; or if the given nmonths represents zero, in which case the
    returned thru date will be the last date of the month of the given from date.

    If only the given thru date is specified then a None from date is returned, UNLESS the given nmonths
    argument represents a negative integer, in which case the returned from date will be nmonths monthss
    previous to the given thru date; or if the given nmonths represents zero, in which case
    the returned from date will be the first date of the month of the given thru date.

    If neither the given from/thru dates are specified then ...
    """
    from_date = _parse_datetime_string(from_date, notz=True)
    thru_date = _parse_datetime_string(thru_date, last_day_of_month_if_no_day=True, notz=True)
    if not isinstance(nmonths, int):
        if isinstance(nmonths, str) and (nmonths := nmonths.strip()):
            try:
                nmonths = int(nmonths)
            except Exception:
                nmonths = 0
        else:
            nmonths = 0
    if from_date:
        if (not thru_date) and isinstance(nmonths, int):
            if nmonths > 0:
                thru_date = add_months(from_date, nmonths)
            elif nmonths == 0:
                thru_date = get_last_date_of_month(from_date)
    elif thru_date:
        if isinstance(nmonths, int):
            if nmonths < 0:
                from_date = add_months(thru_date, nmonths)
            elif nmonths == 0:
                from_date = get_first_date_of_month(thru_date)
    elif isinstance(nmonths, int) and ((nmonths := abs(nmonths)) != 0):
        # If no (valid) from/thru dates given, but the absolute value of nmonths is a non-zero integer, then returns
        # from/thru dates for the last nmonths month ending with the last day of month previous to the current month.
        thru_date = add_months(get_last_date_of_month(), -1)
        from_date = add_months(thru_date, -nmonths)
    if strings is True:
        return (from_date.strftime(f"%Y-%m-%d") if from_date else None,
                thru_date.strftime(f"%Y-%m-%d") if thru_date else None)
    return from_date, thru_date


def get_first_date_of_month(day: Optional[Union[datetime, date, str]] = None) -> datetime:
    """
    Returns a datetime object representing the first day of the month of the given date;
    this given date may be a datetime or date object, or string representing a date or
    datetime; if the given argument is unspecified or incorrect then assumes "today".
    """
    if not (day := _parse_datetime_string(day, notz=True)):
        day = datetime.today().replace(tzinfo=None)
    return day.replace(day=1)


def get_last_date_of_month(day: Optional[Union[datetime, date, str]] = None) -> datetime:
    """
    Returns a datetime object representing the last day of the month of the given date;
    this given date may be a datetime or date object, or string representing a date or
    datetime; if the given argument is unspecified or incorrect then assumes "today".
    """
    if not (day := _parse_datetime_string(day)):
        day = datetime.today().replace(tzinfo=None)
    return datetime(day.year, day.month, calendar.monthrange(day.year, day.month)[1])


def add_months(day: Optional[Union[datetime, date, str]] = None, nmonths: int = 0) -> datetime:
    """
    Returns a datetime object representing the given date with the given nmonths number of months
    added (or substracted if negative) to (or from) that given date.; this given date may be a
    datetime or date object, or string representing a date or datetime; if the given argument
    is unspecified or incorrect then assumes "today".
    """
    if not (day := _parse_datetime_string(day, notz=True)):
        day = datetime.today().replace(tzinfo=None)
    if isinstance(nmonths, int) and (nmonths != 0):
        return day + relativedelta(months=nmonths)
    return day


def _parse_datetime_string(value: Union[str, datetime, date],
                           last_day_of_month_if_no_day: bool = False,
                           notz: bool = False) -> Optional[datetime]:
    """
    Wrapper around dcicutils.datetime_utils.parse_datetime_string to handle a few special cases for convenience.
    """
    last_day_of_month = False
    if not isinstance(value, datetime):
        if isinstance(value, date):
            value = datetime.combine(value, datetime.min.time())
        elif isinstance(value, str):
            if (len(value) == 8) and value.isdigit():
                # Special case to accept for example "20241206" to mean "2024-12-06".
                value = f"{value[0:4]}-{value[4:6]}-{value[6:8]}"
            elif (len(value) == 7) and (value[4] == "-") and value[0:4].isdigit() and value[5:].isdigit():
                # Special case to accept for example "2024-10" to mean "2024-10-01".
                value = f"{value}-01"
                last_day_of_month = last_day_of_month_if_no_day
            elif (len(value) == 7) and (value[2] == "/") and value[0:2].isdigit() and value[3:].isdigit():
                # Special case to accept for example "11/2024" to mean "2024-11-01".
                value = f"{value[3:]}-{value[0:2]}-01"
                last_day_of_month = last_day_of_month_if_no_day
            elif (len(value) == 6) and (value[1] == "/") and value[0:1].isdigit() and value[2:].isdigit():
                # Special case to accept for example "9/2024" to mean "2024-09-01".
                value = f"{value[2:]}-0{value[0:1]}-01"
                last_day_of_month = last_day_of_month_if_no_day
            if not (value := parse_datetime_string(value)):
                return None
        else:
            return None
    value = value.replace(tzinfo=None) if notz is True else value
    if last_day_of_month:
        value = get_last_date_of_month(value)
    return value


import json  # noqa
import os  # noqa
from urllib.parse import unquote  # noqa
from hms_utils.dictionary_utils import get_property, group_items_by_groupings, normalize_elastic_search_aggregation_results  # noqa
from hms_utils.dictionary_print_utils import print_grouped_items  # noqa
from hms_utils.portal.portal_utils import Portal, portal_custom_search  # noqa

if False:
    query_string, aggregation_definition = create_query_for_files(type=DEFAULT_FILE_TYPE,
                                                                  # nmonths=24,
                                                                  aggregation=AGGREGATIONS,
                                                                  max_buckets=AGGREGATION_MAX_BUCKETS,
                                                                  missing_value=AGGREGATION_NO_VALUE)
    print(query_string)
    print(unquote(query_string))
    print(json.dumps(aggregation_definition, indent=4))

    # query = f"/search/?{query_string}&limit=0&from=0"
    query = f"/search/?{query_string}"
    portal = Portal(os.path.expanduser("~/repos/smaht-portal/development.ini"))
    result = portal_custom_search(portal, query, aggregations=aggregation_definition)
    print(json.dumps(result, indent=4))
    result_aggregations = result.get("aggregations")
    result_aggregations_normalized = normalize_elastic_search_aggregation_results(result_aggregations)
    print_grouped_items(result_aggregations_normalized)

if True:
    portal = Portal("smaht-local")
    query = f"/search/?type={DEFAULT_FILE_TYPE}&status={DEFAULT_FILE_STATUS}&data_category%21={DEFAULT_FILE_CATEGORY_IGNORED}"  # noqa
    files = portal.get_metadata(query, limit=1000).get("@graph")
    def map_grouping_value(grouping_name: str, grouping_value: str) -> str:  # noqa
        if grouping_name == AGGREGATION_FIELD_RELEASE_DATE:
            # Hack to turn something like "2024-11-28T04:42:29.531456+00:00" into "2024-11".
            return grouping_value[:7]
        return grouping_value
    def map_grouped_item(grouped_item: Any) -> str:  # noqa
        global files
        for file in files:
            if file.get("uuid") == grouped_item:
                if released_date := _parse_datetime_string(get_property(file, AGGREGATION_FIELD_RELEASE_DATE)):
                    released_date = released_date.strftime(f"%Y-%m-%d")
                    grouped_item += f" ({released_date})"
        return grouped_item
    grouped_files = group_items_by_groupings(files,
                                             groupings=AGGREGATIONS,
                                             sort={AGGREGATION_FIELD_RELEASE_DATE: "-key", None: True},
                                             map_grouping_value=map_grouping_value)
    print(json.dumps(grouped_files, indent=4))
    print_grouped_items(grouped_files, map_grouped_item=map_grouped_item, noitems=True)



from __future__ import annotations
from boto3 import client as BotoClient
from functools import lru_cache
import json
import os
import re
from typing import Any, Callable, Optional, Tuple
from hms_utils.config.config_basic import ConfigBasic
from hms_utils.dictionary_parented import JSON
from hms_utils.chars import chars
from hms_utils.config.config_with_secrets import ConfigWithSecrets
from hms_utils.env_utils import os_environ


class ConfigWithAwsMacros(ConfigBasic):

    _AWS_SECRET_MACRO_NAME_PREFIX = "aws-secret:"
    _AWS_SECRET_MACRO_START = f"{ConfigBasic._MACRO_START}{_AWS_SECRET_MACRO_NAME_PREFIX}"
    _AWS_SECRET_MACRO_END = ConfigBasic._MACRO_END
    _AWS_SECRET_MACRO_PATTERN = re.compile(r"\$\{aws-secret:([^}]+)\}")
    _AWS_SECRET_NAME_NAME = "IDENTITY"
    _AWS_PROFILE_ENV_NAME = "AWS_PROFILE"
    _AWS_CACHED_ACCOUNT_NUMBERS = {}
    _TYPE_NAME_AWS = "aws"

    def __init__(self, config: JSON,
                 name: Optional[str] = None,
                 path_separator: Optional[str] = None,
                 custom_macro_lookup: Optional[Callable] = None,
                 raise_exception: bool = False,
                 aws_secrets_name: Optional[str] = None,
                 noaws: bool = False, **kwargs) -> None:
        self._aws_secrets_name = aws_secrets_name.strip() if isinstance(aws_secrets_name, str) else None
        self._noaws = noaws is True
        self._raise_exception = raise_exception is True
        super().__init__(config,
                         name=name,
                         path_separator=path_separator,
                         custom_macro_lookup=self._lookup_macro_custom,
                         raise_exception=raise_exception, **kwargs)

    @property
    def aws_secrets_name(self) -> Optional[str]:
        return self._aws_secrets_name

    @aws_secrets_name.setter
    def aws_secrets_name(self, value: str) -> Optional[str]:
        self._aws_secrets_name = value.strip() if isinstance(value, str) else None

    def _lookup_macro_custom(self, macro_value: str, context: Optional[JSON] = None) -> Any:
        if not macro_value.startswith(ConfigWithAwsMacros._AWS_SECRET_MACRO_NAME_PREFIX):
            return None
        secret_specifier = macro_value[len(ConfigWithAwsMacros._AWS_SECRET_MACRO_NAME_PREFIX):]
        return self._lookup_aws_secret(secret_specifier, context)

    def _lookup_aws_secret(self, secret_specifier: str, context: Optional[JSON] = None) -> Optional[str]:
        def lookup_environment_variable(name: str, context: Optional[JSON] = None) -> Optional[str]:
            nonlocal self
            if aws_profile := self.lookup(name, context=context, show=True):
                return aws_profile
            return os.environ.get(name)
            if isinstance(context, JSON):
                while True:
                    if aws_profile := context.get(name):
                        return aws_profile
                    if not (context := context.parent):
                        break
            return os.environ.get(name)
        aws_profile = None ; secrets_name = None ; secret_name = None  # noqa
        if self._path_separator in secret_specifier:
            if split_secret_specifier := secret_specifier.split(self._path_separator):
                if len(split_secret_specifier) == 3:
                    aws_profile = split_secret_specifier[0]
                    secrets_name = split_secret_specifier[1]
                    secret_name = split_secret_specifier[2]
                elif len(split_secret_specifier) == 2:
                    secrets_name = split_secret_specifier[0]
                    secret_name = split_secret_specifier[1]
        if (not secrets_name) or (not secret_name):
            secret_name = secret_specifier
            if self._aws_secrets_name:
                secrets_name = self._aws_secrets_name
            else:
                secrets_name = lookup_environment_variable(ConfigWithAwsMacros._AWS_SECRET_NAME_NAME, context)
        if not (secret_name and secrets_name):
            return None
        if aws_profile:
            with os_environ(ConfigWithAwsMacros._AWS_PROFILE_ENV_NAME, aws_profile):
                return self._aws_get_secret(secrets_name, secret_name, aws_profile)
        if aws_profile := lookup_environment_variable(ConfigWithAwsMacros._AWS_PROFILE_ENV_NAME, context):
            with os_environ(ConfigWithAwsMacros._AWS_PROFILE_ENV_NAME, aws_profile):
                return self._aws_get_secret(secrets_name, secret_name, aws_profile)
        return self._aws_get_secret(secrets_name, secret_name, aws_profile)

    def _aws_get_secret(self, secrets_name: str, secret_name: str, aws_profile: Optional[str]) -> Optional[str]:
        if self._noaws:
            return None
        value, account_number = self._aws_read_secret(secrets_name, secret_name, aws_profile)
        if value is not None:
            if isinstance(self, ConfigWithSecrets) and self.secrets:
                # See: ConfigWithAwsMacros._secrets_plaintext_value
                value = self._secrets_encoded(
                    f"{account_number}{ConfigWithSecrets._TYPE_NAME_SEPARATOR}"
                    f"{secrets_name}{ConfigWithSecrets._TYPE_NAME_SEPARATOR}"
                    f"{secret_name}{ConfigWithSecrets._TYPE_NAME_SEPARATOR}{value}",
                    value_type=ConfigWithAwsMacros._TYPE_NAME_AWS)
        return value

    @lru_cache
    def _aws_read_secret(self, secrets_name: str, secret_name: str,
                         aws_profile: Optional[str]) -> Tuple[Optional[str], Optional[str]]:
        try:
            self._debug(f"Reading AWS secret {secrets_name}/{secret_name}"
                        f"{f' {chars.dot} profile: {aws_profile}' if aws_profile else ''}")
            secrets, account_number = self._aws_read_secrets(secrets_name, aws_profile)
            if (not secrets) or ((value := secrets.get(secret_name)) is None):
                self._warning(f"Cannot find AWS secret {secrets_name}/{secret_name}"
                              f"{f' {chars.dot} profile: {aws_profile}' if aws_profile else ''}")
                return None, None
            self._debug(f"Read AWS secret OK: {secrets_name}/{secret_name}"
                        f"{f' {chars.dot} profile: {aws_profile}' if aws_profile else ''}")
            return value, account_number
        except Exception as e:
            if self._raise_exception is True:
                raise e
            msg = self._aws_error_message(f"Cannot read AWS secret {secrets_name}/{secret_name}", aws_profile, e)
            self._debug(msg)
            self._warning(msg)
            return None, None

    @lru_cache
    def _aws_read_secrets(self, secrets_name: str, aws_profile: Optional[str]) -> Tuple[Optional[str], Optional[str]]:
        def extract_aws_account_number(secrets: dict) -> str:
            try:
                return secrets["ARN"].split(':')[4].strip()
            except Exception:
                return ""
        try:
            self._debug(f"Reading AWS secrets {secrets_name}"
                        f"{f' {chars.dot} profile: {aws_profile}' if aws_profile else ''}")
            boto_secrets = ConfigWithAwsMacros._boto_client("secretsmanager")
            secrets = boto_secrets.get_secret_value(SecretId=secrets_name)
            account_number = extract_aws_account_number(secrets)
            secrets = json.loads(secrets.get("SecretString"))
            self._debug(lambda: self._aws_error_message(f"Read AWS secrets OK: {secrets_name}", aws_profile))
            return secrets, account_number
        except Exception as e:
            if self._raise_exception is True:
                raise e
            msg = self._aws_error_message(f"Cannot read AWS secrets {secrets_name}", aws_profile, e)
            self._debug(msg)
            self._warning(msg)
            return None, None

    def _aws_error_message(self, message: str, aws_profile: Optional[str],
                           exception: Optional[Exception] = None, _noprofile: bool = False) -> str:
        if _noprofile is not True:
            if aws_profile:
                message += f" {chars.dot} profile: {aws_profile}"
            if aws_account_number := self._aws_current_account_number(aws_profile):
                message += f" {chars.dot} account: {aws_account_number}"
            elif not aws_profile:
                message += f" {chars.dot} profile: unspecified"
        if exception := str(exception):
            if ("token" in exception) and ("expired" in exception):
                message += f" {chars.dot} expired"
            elif ("not" in exception) and ("found" in exception):
                message += f" {chars.dot} unknown"
        return message

    @lru_cache
    def _aws_current_account_number(self, aws_profile: Optional[str]) -> Optional[str]:
        try:
            self._debug(f"Reading AWS account number{f': {aws_profile}' if aws_profile else ''}")
            aws_account_number = self._boto_client("sts").get_caller_identity()["Account"]
            self._debug(f"Read AWS account number OK{f': {aws_profile}' if aws_profile else ''}")
            return aws_account_number
        except Exception as e:
            msg = self._aws_error_message(
                f"Cannot read AWS account number{f': {aws_profile}' if aws_profile else ''}",
                aws_profile, e, _noprofile=True)
            self._debug(msg)
            return None

    def _contains_aws_secret_values(self, value: Any) -> bool:
        if not isinstance(value, str):
            return False
        if (start := value.find(ConfigWithSecrets._SECRET_VALUE_START + ConfigWithAwsMacros._TYPE_NAME_AWS)) < 0:
            return False
        return value.find(ConfigWithSecrets._SECRET_VALUE_END) > start

    def _secrets_plaintext_value(self, value: str) -> Optional[str]:
        # See: ConfigWithAwsMacros._aws_get_secret
        if (value.startswith(ConfigWithAwsMacros._TYPE_NAME_AWS) and
            (len(value_parts := value.split(ConfigWithSecrets._TYPE_NAME_SEPARATOR)) >= 5)):  # noqa
            return ConfigWithSecrets._TYPE_NAME_SEPARATOR.join(value_parts[4:])
        return None

    def _secrets_plaintext_info(self, value: str) -> Tuple[Optional[str], Optional[str], Optional[str]]:
        if (value.startswith(ConfigWithSecrets._SECRET_VALUE_START) and
            value.endswith(ConfigWithSecrets._SECRET_VALUE_END)):  # noqa
            value = value[ConfigWithSecrets._SECRET_VALUE_START_LENGTH:-ConfigWithSecrets._SECRET_VALUE_END_LENGTH]
            if ((len(secrets_encoded_parts := value.split(ConfigWithSecrets._TYPE_NAME_SEPARATOR)) >= 5) and
                secrets_encoded_parts[4]):  # noqa
                if ((aws_account_number := secrets_encoded_parts[1]) and
                    (aws_secrets_name := secrets_encoded_parts[2]) and
                    (aws_secret_name := secrets_encoded_parts[3])):  # noqa
                    return aws_account_number, aws_secrets_name, aws_secret_name
        return None, None, None

    def _note_macro_not_found(self, macro_value: str,
                              context: Optional[JSON] = JSON, context_path: Optional[str] = None) -> None:
        if not macro_value.startswith(ConfigWithAwsMacros._AWS_SECRET_MACRO_NAME_PREFIX):
            super()._note_macro_not_found(macro_value, context, context_path=context_path)

    @staticmethod
    def _boto_client(service: str) -> object:
        # This boto3.DEFAULT_SESSION works around an oddity discovered way back (circa 2022-06-19) with the
        # boto3 session getting stuck/cached or something once used once and then again with a different profile.
        # https://hms-dbmi.slack.com/archives/D03ENS13XA7/p1655648553779689
        import boto3
        boto3.DEFAULT_SESSION = None
        return BotoClient(service)



# Convenience utility to maintain JSON based configuration properties.

from __future__ import annotations
from boto3 import client as BotoClient
from copy import deepcopy
from functools import lru_cache
import io
import json
import os
import re
import stat
import sys
import traceback
from typing import Any, List, Optional, Tuple, Union
import yaml
from hms_utils.chars import chars
from hms_utils.dictionary_utils import JSON, delete_paths_from_dictionary, sort_dictionary
from hms_utils.dictionary_print_utils import print_dictionary_list, print_dictionary_tree
from hms_utils.version_utils import get_version
from hms_utils.terminal_utils import terminal_color as color

DEFAULT_CONFIG_DIR = "~/.config/hms"
DEFAULT_CONFIG_FILE_NAME = "config.json"
DEFAULT_SECRETS_FILE_NAME = "secrets.json"
DEFAULT_PATH_SEPARATOR = "/"
DEFAULT_EXPORT_NAME_SEPARATOR = ":"
AWS_PROFILE_ENV_NAME = "AWS_PROFILE"
OBFUSCATED_VALUE = "********"

SUPPRESS_AWS_SECRET_NOT_FOUND_WARNING = False  # Hack


def main():

    args = parse_args(sys.argv[1:])

    # TODO
    # After implementing the "import" thing I realized i *thjink* it could be generalized
    # such that secrets couild be treated like this rather than as treating them specially.

    config = None
    if args.config_file:
        try:
            config = Config(args.config_file,
                            config_imports=args.config_imports,
                            secrets_imports=args.secrets_imports,
                            path_separator=args.path_separator,
                            nomacros=args.nomacros, noaws=args.noaws)
        except Exception as e:
            error(f"Cannot process config file: {args.config_file}", exception=e, trace=True)

    secrets = None
    if args.secrets_file:
        try:
            secrets = Config(args.secrets_file,
                             path_separator=args.path_separator,
                             nomacros=args.nomacros, noaws=args.noaws)
        except Exception as e:
            error(f"Cannot process secrets file: {args.secrets_file}", exception=e, trace=True)

    merged_config = config.merge_secrets(secrets) if secrets else config

    if not args.names:
        if (not args.nomerge) and (not args.json) and (not args.yaml) and config and secrets:
            print_config_and_secrets_merged(merged_config, args)
        else:
            print_config_and_secrets_unmerged(config, secrets, args)
        exit(0)

    def setup_aws_profile_environment_variable():
        # TOTAL HACK (in a hurry 2024-10-08). This does what below args.exports
        # loop does but just for AWS_PROFILE environment variable setting.
        nonlocal args, merged_config
        for name in args.names:
            if (value := merged_config.lookup(name, allow_dictionary=True, raw=True)) is not None:
                if (name == AWS_PROFILE_ENV_NAME) and (AWS_PROFILE_ENV_NAME not in os.environ):
                    # Special case to handle list of paths the first of which specifies AWS_PROFILE,
                    # and which needs to be set to evaluate subsequent paths which are aws-secret macro values.
                    os.environ[AWS_PROFILE_ENV_NAME] = value
                # TODO: Refactor this increasingtly unwieldy logic.
                if isinstance(value, dict):
                    # Special case: If target name/path is a dictionary then generate
                    # exports for every direct (non-dictionary) key/value within it.
                    for key in value:
                        if ((single_value := value[key]) is not None) and (not isinstance(single_value, dict)):
                            if (key == AWS_PROFILE_ENV_NAME) and (AWS_PROFILE_ENV_NAME not in os.environ):
                                # Same special case as above for (direct) items within a dictionary.
                                os.environ[AWS_PROFILE_ENV_NAME] = single_value
                    if True:
                        # Walk up the hierarchy to get direct values of each parent/ancestor.
                        parent = value.parent
                        while parent:
                            for key in parent:
                                if ((single_value := parent[key]) is not None) and (not isinstance(single_value, dict)):
                                    if (key == AWS_PROFILE_ENV_NAME) and (AWS_PROFILE_ENV_NAME not in os.environ):
                                        os.environ[AWS_PROFILE_ENV_NAME] = single_value
                            parent = parent.parent

    status = 0
    if args.export:
        if args.export_file and os.path.exists(args.export_file):
            error(f"Export file must not already exist: {args.export_file}")
        exports = {}
        setup_aws_profile_environment_variable()
        for name in args.names:
            if (colon := name.find(DEFAULT_EXPORT_NAME_SEPARATOR)) > 0:
                export_name = name[0:colon]
                if not (name := name[colon + 1:].strip()):
                    continue
            else:
                export_name = path_basename(name, args.path_separator)
            found = False ; found_dictionary = False  # noqa
            if (value := merged_config.lookup(name, allow_dictionary=True, raw=True)) is not None:
                if (export_name == AWS_PROFILE_ENV_NAME) and (AWS_PROFILE_ENV_NAME not in os.environ):
                    # Special case to handle list of paths the first of which specifies AWS_PROFILE,
                    # and which needs to be set to evaluate subsequent paths which are aws-secret macro values.
                    os.environ[AWS_PROFILE_ENV_NAME] = value
                # TODO: Refactor this increasingtly unwieldy logic.
                if isinstance(value, dict):
                    # Special case: If target name/path is a dictionary then generate
                    # exports for every direct (non-dictionary) key/value within it.
                    found_dictionary = True
                    for key in value:
                        if ((single_value := value[key]) is not None) and (not isinstance(single_value, dict)):
                            if (key == AWS_PROFILE_ENV_NAME) and (AWS_PROFILE_ENV_NAME not in os.environ):
                                # Same special case as above for (direct) items within a dictionary.
                                os.environ[AWS_PROFILE_ENV_NAME] = single_value
                            # TODO: A bit shaky on this ...
                            if merged_config._contains_macro(single_value):
                                single_value = merged_config._expand_macro_value(single_value, value)
                            # TODO: A bit shaky on this ...
                            if merged_config._contains_aws_secret_macro(single_value):
                                # Note the trailing separator/slash on the context.
                                aws_secret_context_path = f"{name}{args.path_separator}"
                                single_value = config._expand_aws_secret_macros(
                                    single_value, aws_secret_context_path=aws_secret_context_path)
                            exports[key] = single_value
                            found = True
                    if True:
                        # Walk up the hierarchy to get direct values of each parent/ancestor.
                        parent = value.parent
                        while parent:
                            for key in parent:
                                if ((single_value := parent[key]) is not None) and (not isinstance(single_value, dict)):
                                    if (key == AWS_PROFILE_ENV_NAME) and (AWS_PROFILE_ENV_NAME not in os.environ):
                                        os.environ[AWS_PROFILE_ENV_NAME] = single_value
                                    # TODO: Shaky on this ...
                                    if merged_config._contains_macro(single_value):
                                        single_value = merged_config._expand_macro_value(single_value, value)
                                    # TODO: Shaky on this ...
                                    if merged_config._contains_aws_secret_macro(single_value):
                                        # Note the trailing separator/slash on the context.
                                        aws_secret_context_path = f"{name}{args.path_separator}"
                                        single_value = config._expand_aws_secret_macros(
                                            single_value, aws_secret_context_path=aws_secret_context_path)
                                    if key not in exports:
                                        exports[key] = single_value
                            parent = parent.parent
                else:
                    exports[export_name] = value
                    found = True
            if not found:
                if args.nowarning:
                    if found_dictionary:
                        warning(f"{chars.rarrow} Config name/path contains no direct values: {name}")
                    else:
                        warning(f"{chars.rarrow} Config name/path not found: {name}")
                status = 1
        exports = dict(sorted(exports.items()))
        if args.export_file:
            if args.verbose:
                print(f"Writing exports to file: {args.export_file}")
            with io.open(args.export_file, "w") as f:
                for export in exports:
                    export = f"export {export}={exports[export]}"
                    if args.verbose:
                        print(f"{chars.rarrow_hollow} {export}")
                    f.write(f"{export}\n")
        else:
            for export in sorted(exports):
                export = f"export {export}={exports[export]}"
                print(export)
    else:
        for name in args.names:
            if merged_config and ((value := merged_config.lookup(
                                            name, allow_dictionary=args.json,
                                            aws_secret_context_path=args.aws_secret_context_path)) is not None):
                if args.json_formatted and isinstance(value, dict):
                    print(json.dumps(value, indent=4))
                else:
                    if args.verbose:
                        print(f"{name}: {value}")
                    else:
                        print(value)
            else:
                if not args.nowarning:
                    warning(f"Cannot find name/path: {name}")
                status = 1

    sys.exit(status)


def main_show_script_path():
    sys.argv = ["hmx-config", "--functions"]
    main()


def parse_args(argv: List[str]) -> object:

    class Args:
        config_dir = os.path.expanduser(DEFAULT_CONFIG_DIR)
        config_dir_explicit = False
        config_file = DEFAULT_CONFIG_FILE_NAME
        config_file_explicit = False
        config_imports = []
        import_config_files = []
        secrets_file = DEFAULT_SECRETS_FILE_NAME
        secrets_file_explicit = False
        secrets_imports = []
        aws_secret_context_path = None
        import_secrets_files = []
        path_separator = DEFAULT_PATH_SEPARATOR
        name = None
        names = []
        list = False
        yaml = False
        json = False
        json_formatted = False
        json_only = False
        raw = False
        show_secrets = False
        show_paths = False
        nocolor = False
        nomerge = False
        nosort = False
        nomacros = False
        export = False
        export_file = None
        noaws = False
        verbose = False
        nowarning = False
        debug = False

    args = Args()

    if value := os.environ.get("HMS_CONFIG_DIR"):
        args.config_dir = value
        args.config_dir_explicit = True
    if value := os.environ.get("HMS_CONFIG"):
        args.config_file = value
        args.config_file_explicit = True
    if value := os.environ.get("HMS_SECRETS"):
        args.secrets_file = value
        args.secrets_file_explicit = True

    argi = 0 ; argn = len(argv)  # noqa
    while argi < argn:
        arg = argv[argi] ; argi += 1  # noqa
        if arg in ["--functions", "-functions", "--function", "-function", "--shell", "-shell"]:
            print(os.path.join(os.path.dirname(os.path.abspath(__file__)), "hms_config.sh"))
            exit(0)
        elif arg in ["--dir", "-dir", "--directory", "-directory"]:
            if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                _usage()
            args.config_dir = arg ; args.config_dir_explicit = True ; argi += 1  # noqa
        elif arg in ["--config", "-config", "--conf", "-conf"]:
            if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                _usage()
            args.config_file = arg
            args.config_file_explicit = True
            argi += 1
        elif arg in ["--secrets-config", "-secrets-config", "--secrets-conf", "-secrets-conf", "--secret-config",
                     "-secret-config", "--secret-conf", "-secret-conf", "--secrets", "-secrets", "--secret", "-secret"]:
            if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                _usage()
            args.secrets_file = arg
            args.secrets_file_explicit = True
            argi += 1
        elif arg in ["--imports", "-imports", "--import", "-import"]:
            if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                _usage()
            if "secret" in args.lower():
                args.import_secret_files.append(arg)
            else:
                args.import_config_files.append(arg)
            argi += 1
        elif arg in ["--import-config", "-import-config", "--import-conf", "-import-conf",
                     "--iconfig", "-iconfig", "--iconf", "-iconf"]:
            if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                _usage()
            args.import_config_files.append(arg)
            argi += 1
        elif arg in ["--import-secrets", "-import-secrets", "--import-secret", "-import-secret",
                     "--isecrets", "-isecrets", "--isecret", "-isecret"]:
            if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                _usage()
            args.import_secrets_files.append(arg)
            argi += 1
        elif arg in ["--context", "-context"]:
            if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                _usage()
            args.aws_secret_context_path = arg
            argi += 1
        elif arg in ["--path-separator", "-path-separator", "--separator", "-separator", "--sep", "-sep"]:
            if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                _usage()
            args.path_separator = arg ; argi += 1  # noqa
        elif arg in ["--show-secrets", "-show-secrets", "--show-secret", "-show-secret", "--show", "-show"]:
            args.show_secrets = True
        elif arg in ["--show-paths", "-show-paths", "--show-path",
                     "-show-path", "--paths", "-paths", "--path", "-path"]:
            args.show_paths = True
        elif arg in ["--list", "-list"]:
            args.list = True
        elif arg in ["--yaml", "-yaml", "--yml", "-yml"]:
            args.yaml = True
        elif arg in ["--json", "-json"]:
            args.json = True
        elif arg in ["--json-only", "-json-only", "--jsononly", "-jsononly"]:
            args.json = True
            args.json_only = True
        elif arg in ["--jsonf", "-jsonf", "--json-formatted", "-json-formatted", "--json-format", "-json-format"]:
            args.json = True
            args.json_formatted = True
        elif arg in ["--raw", "-raw"]:
            args.raw = True
        elif arg in ["--nocolor", "-nocolor"]:
            args.nocolor = True
        elif arg in ["--nomerge", "-nomerge"]:
            args.nomerge = True
        elif arg in ["--nosort", "-nosort"]:
            args.nosort = True
        elif arg in ["--nomacros", "-nomacros", "--nomacro", "-nomacro"]:
            args.nomacros = True
        elif arg in ["--export", "-export", "--exports", "-exports"]:
            args.export = True
        elif arg in ["--export-file", "-export-file"]:
            if (argi >= argn) or not (arg := argv[argi]) or (not arg):
                _usage()
            args.export = True
            args.export_file = argv[argi] ; argi += 1  # noqa
        elif arg in ["-noaws", "noaws"]:
            args.noaws = True
        elif arg in ["--debug", "-debug"]:
            args.debug = True
        elif arg in ["--nowarning", "-nowarning", "--nowarnings", "-nowarnings"]:
            args.nowarning = True
        elif arg in ["--verbose", "-verbose"]:
            args.verbose = True
        elif arg in ["--version", "-version"]:
            print(f"hmx-utils version: {get_version()}")
            exit(0)
        elif (arg in ["--help", "-help"]) or arg.startswith("-"):
            _usage()
        else:
            args.names.append(arg)

    # If either config or secrets file is explicit ignore the other implicit one.
    if args.config_file_explicit:
        if not args.secrets_file_explicit:
            args.secrets_file = None
    elif args.secrets_file_explicit:
        args.config_file = None

    if args.names and (args.show_secrets or args.show_paths or
                       args.yaml or args.nosort or args.nomerge or
                       args.nomacros or args.nocolor or args.yaml or args.list):
        error("Option not allowed with a config name/path argument.", usage=True)
    elif args.export and (not args.names):
        error("The --export option must be used with a config name/path argument.", usage=True)

    config_file, secrets_file = resolve_files(args)
    if (not config_file) and (not secrets_file):
        error("No config or secrets file found.", usage=True)
    args.config_file = config_file
    args.secrets_file = secrets_file

    # TODO: Refactor with other file resolution code; and with yaml support etc.
    for import_config_file in args.import_config_files:
        if import_config_file.lower() == "default":
            import_config_file = args.config_file or DEFAULT_CONFIG_FILE_NAME
        import_config_file = resolve_file_path(import_config_file, args.config_dir, file_explicit=True,
                                               directory_explicit=args.config_dir_explicit)
        if not import_config_file:
            error(f"Cannot open (import) config file: {import_config_file}")
        try:
            with io.open(import_config_file) as f:
                if import_config_file.startswith(".yaml") or import_config_file.startswith(".yml"):
                    config_import_json = JSON(yaml.safe_load(f))
                else:
                    config_import_json = JSON(json.load(f))
        except Exception:
            error(f"Cannot load (import) config file: {import_config_file}")
        args.config_imports.append(config_import_json)

    for import_secrets_file in args.import_secrets_files:
        if import_secrets_file.lower() == "default":
            import_secrets_file = args.secrets_file or DEFAULT_SECRETS_FILE_NAME
        import_secrets_file = resolve_file_path(import_secrets_file, args.config_dir, file_explicit=True,
                                                directory_explicit=args.config_dir_explicit)
        if not os.path.exists(import_secrets_file):
            error(f"Cannot open (import) secrets file: {import_secrets_file}")
        try:
            with io.open(import_secrets_file) as f:
                if import_secrets_file.startswith(".yaml") or import_secrets_file.startswith(".yml"):
                    secrets_import_json = yaml.safe_load(f)
                else:
                    secrets_import_json = JSON(json.load(f))
        except Exception:
            error(f"Cannot load (import) secrets file: {import_secrets_file}")
        args.secrets_imports.append(secrets_import_json)

    return args


# def tree_key_modifier(secrets: Config, args: object, key_path: str, key: Optional[str] = None) -> Optional[str]:
#     return ((key or key_path) if ((not secrets) or (secrets.lookup(key_path) is None))
#             else color(key or key_path, "red", nocolor=args.nocolor))


# def tree_value_modifier(secrets: Config, args: object, key_path: str, value: str) -> Optional[str]:
#     if (not args.show_secrets) and secrets and secrets.contains(key_path):
#         value = OBFUSCATED_VALUE
#     return value if ((not secrets) or
#                      (secrets.lookup(key_path) is None)) else color(value, "red", nocolor=args.nocolor)


def print_config_and_secrets_merged(merged_config: Config, args: object) -> None:

    global SUPPRESS_AWS_SECRET_NOT_FOUND_WARNING
    SUPPRESS_AWS_SECRET_NOT_FOUND_WARNING = True
    merged_config_json = merged_config.json
    secrets = merged_config.secrets if hasattr(merged_config, "secrets") else None
    unmerged_secrets = merged_config.unmerged_secrets if hasattr(merged_config, "unmerged_secrets") else None
    merged_secrets = merged_config.merged_secrets if hasattr(merged_config, "merged_secrets") else None

    def tree_key_modifier(key_path: str, key: Optional[str] = None) -> Optional[str]:
        nonlocal args, secrets
        return ((key or key_path) if ((not secrets) or (not secrets.contains(key_path)))
                else color(key or key_path, "red", nocolor=args.nocolor))

    def tree_value_modifier(key_path: str, value: str) -> Optional[str]:
        nonlocal args, secrets
        if (not args.show_secrets) and secrets and secrets.contains(key_path):
            value = OBFUSCATED_VALUE
        return value if ((not secrets) or
                         (not secrets.contains(key_path))) else color(value, "red", nocolor=args.nocolor)

    def tree_value_annotator(key_path: str) -> Optional[str]:
        nonlocal unmerged_secrets
        if unmerged_secrets and (key_path in unmerged_secrets):
            return f"{chars.rarrow} unmerged {chars.xmark}"
        return None

    def tree_value_annotator_secrets(key_path: str) -> Optional[str]:
        nonlocal merged_secrets, unmerged_secrets
        if unmerged_secrets and (key_path in unmerged_secrets):
            return f"{chars.rarrow} unmerged {chars.xmark}"
        elif merged_secrets and (key_path in merged_secrets):
            return f"{chars.rarrow_hollow} merged {chars.check}"
        return None

    def tree_arrow_indicator(key_path: str, value: Any) -> str:
        nonlocal secrets
        return chars.rarrow_hollow if (secrets and secrets.contains(key_path)) else ""

    if merged_config and secrets:
        if not args.nosort:
            merged_config_json = sort_dictionary(merged_config_json)
        print(f"\n{merged_config.file}: [with {os.path.basename(args.secrets_file)}"
              f"{' partially' if unmerged_secrets else ''} merged in]")
        if args.list:
            print_dictionary_list(merged_config_json, path_separator=args.path_separator,
                                  prefix=f" {chars.rarrow_hollow} ",
                                  key_modifier=tree_key_modifier,
                                  value_modifier=tree_value_modifier,
                                  value_annotator=tree_value_annotator)
        else:
            print_dictionary_tree(merged_config_json, indent=1, paths=args.show_paths,
                                  path_separator=args.path_separator,
                                  key_modifier=tree_key_modifier,
                                  value_modifier=tree_value_modifier,
                                  value_annotator=tree_value_annotator,
                                  arrow_indicator=tree_arrow_indicator)
        if unmerged_secrets:
            print(f"\n{secrets.file}: [secrets unmerged]")
            secrets_json = delete_paths_from_dictionary(secrets.json, merged_secrets)
            if args.list:
                print_dictionary_list(secrets_json, path_separator=args.path_separator,
                                      prefix=f" {chars.rarrow_hollow} ",
                                      key_modifier=tree_key_modifier,
                                      value_modifier=tree_value_modifier,
                                      value_annotator=tree_value_annotator)
            else:
                print_dictionary_tree(secrets_json, indent=1, paths=args.show_paths, path_separator=args.path_separator,
                                      key_modifier=tree_key_modifier,
                                      value_modifier=tree_value_modifier,
                                      value_annotator=tree_value_annotator_secrets,
                                      arrow_indicator=tree_arrow_indicator)
            if args.debug:
                print("\nMerged from secrets:") ; [print(f"- {item}") for item in merged_secrets]  # noqa
                print("\nUnmerged from secrets") ; [print(f"- {item}") for item in unmerged_secrets]  # noqa
        print()
    SUPPRESS_AWS_SECRET_NOT_FOUND_WARNING = False


def print_config_and_secrets_unmerged(config: Config, secrets: Config, args: object) -> None:

    secrets_imports = config.secrets_imports

    def tree_key_modifier(key_path: str, key: Optional[str] = None) -> Optional[str]:
        nonlocal args, secrets, secrets_imports
        secret = False
        if key_path:
            if secrets and secrets.contains(key_path):
                secret = True
        return color(key or key_path, "red", nocolor=args.nocolor) if secret else (key or key_path)

    def tree_value_modifier(key_path: str, value: str) -> Optional[str]:
        nonlocal args, secrets
        if (not args.show_secrets) and secrets and secrets.contains(key_path):
            value = OBFUSCATED_VALUE
        return value if ((not secrets) or
                         (not secrets.contains(key_path))) else color(value, "red", nocolor=args.nocolor)

    if config:
        if not args.json_only:
            print(f"\n{config.file}:")
        data = config.json if not (args.debug or args.raw) else config.rawjson
        if not args.nosort:
            data = sort_dictionary(data)
        if args.yaml:
            print(yaml.dump(data))
        elif args.json:
            if True:  # TODO: This seem very dicey; added for 4dn-cloud-infra secrets.json generation.
                for key in data:
                    if Config._is_primitive_type(value := data[key]) and config._contains_aws_secret_macro(value):
                        if (value := config._expand_aws_secret_macros(value, aws_secret_context_path=None)) is not None:
                            data[key] = value
            print(json.dumps(data, indent=4))
        elif args.list:
            print_dictionary_list(data, path_separator=args.path_separator, prefix=f" {chars.rarrow_hollow} ")
        else:
            print_dictionary_tree(data, indent=1, paths=args.show_paths, path_separator=args.path_separator,
                                  key_modifier=tree_key_modifier,
                                  value_modifier=tree_value_modifier)
    if secrets:
        if not args.json_only:
            print(f"\n{secrets.file}:")
        data = secrets.json if not (args.debug or args.raw) else secrets.rawjson
        if args.yaml:
            print(yaml.dump(data))
        elif args.json:
            print(json.dumps(data, indent=4))
        else:
            if args.list:
                print_dictionary_list(data, path_separator=args.path_separator, prefix=f" {chars.rarrow_hollow} ",
                                      key_modifier=tree_key_modifier,
                                      value_modifier=tree_value_modifier)
            else:
                print_dictionary_tree(
                    data, indent=1,
                    paths=args.show_paths, path_separator=args.path_separator,
                    key_modifier=tree_key_modifier,
                    value_modifier=tree_value_modifier)
    if not args.json_only:
        print()


def resolve_files(args: List[str]) -> Tuple[Optional[str], Optional[str]]:

    def ensure_secrets_file_protected(file: str) -> bool:
        try:
            return stat.S_IMODE(os.stat(file).st_mode) in (0o600, 0o400)
        except Exception:
            return False

    config_file = args.config_file
    secrets_file = args.secrets_file

    if config_file:
        if not (config_file := resolve_file_path(config_file, args.config_dir,
                                                 file_explicit=args.config_file_explicit,
                                                 directory_explicit=args.config_dir_explicit)):
            if args.config_file.endswith(".json"):
                config_file = args.config_file[:-5] + ".yaml"
                config_file = resolve_file_path(config_file, args.config_dir,
                                                file_explicit=args.config_file_explicit,
                                                directory_explicit=args.config_dir_explicit)
        if (not config_file) and args.config_file_explicit:
            error(f"Config file not found: {args.config_file}")

    if secrets_file:
        if not (secrets_file := resolve_file_path(secrets_file, args.config_dir,
                                                  file_explicit=args.secrets_file_explicit,
                                                  directory_explicit=args.config_dir_explicit)):
            if args.secrets_file.endswith(".json"):
                secrets_file = args.secrets_file[:-5] + ".yaml"
                secrets_file = resolve_file_path(secrets_file, args.config_dir,
                                                 file_explicit=args.secrets_file_explicit,
                                                 directory_explicit=args.config_dir_explicit)
        if (not secrets_file) and args.secrets_file_explicit:
            error(f"Config secrets file not found: {args.config_file_explicit}")
        if not ensure_secrets_file_protected(secrets_file):
            warning(f"Your secrets file is not read protected from others: {secrets_file}")

    return config_file, secrets_file


def resolve_file_path(file: str, directory: Optional[str] = None,
                      file_explicit: bool = False, directory_explicit: bool = False) -> Optional[str]:
    if os.path.isabs(file := os.path.normpath(file or "")):
        if not os.path.exists(file):
            return None
        return file
    elif file_explicit and os.path.exists(file) and (not os.path.isdir(file)) and (not directory_explicit):
        if not os.path.isabs(file):
            file = os.path.join(os.path.abspath(os.curdir), file)
        return file
    elif os.path.isabs(directory := os.path.normpath(directory or "")):
        if not os.path.isdir(directory):
            return None
    if not os.path.isdir(directory := os.path.normpath(os.path.join(os.path.abspath(os.curdir), directory))):
        return None
    elif not os.path.exists(file := os.path.join(directory, file)):
        return None
    return file


def path_basename(name: str, separator: str = DEFAULT_PATH_SEPARATOR) -> str:
    if (index := name.rfind(separator)) > 0:
        return name[index + 1:]
    return name


class Config:

    _MACRO_START = "${"
    _MACRO_END = "}"
    _MACRO_PATTERN = re.compile(r"\$\{([^}]+)\}")
    _MACRO_HIDE_START = "@@@__["
    _MACRO_HIDE_END = "]__@@@"
    _MACRO_PATTERN = re.compile(r"\$\{([^}]+)\}")
    _AWS_SECRET_MACRO_NAME_PREFIX = "aws-secret:"
    _AWS_SECRET_MACRO_START = f"{_MACRO_START}{_AWS_SECRET_MACRO_NAME_PREFIX}"
    _AWS_SECRET_MACRO_END = _MACRO_END
    _AWS_SECRET_MACRO_PATTERN = re.compile(r"\$\{aws-secret:([^}]+)\}")
    _AWS_SECRET_NAME_NAME = "IDENTITY"
    _IMPORTED_CONFIG_KEY_PREFIX = "@@@__CONFIG__:"
    _IMPORTED_SECRETS_KEY_PREFIX = "@@@__SECRETS__:"

    def __init__(self, file_or_dictionary: Union[str, dict],
                 config_imports: List[dict] = [], secrets_imports: List[dict] = [],
                 path_separator: str = DEFAULT_PATH_SEPARATOR, nomacros: bool = False, noaws: bool = False) -> None:
        self._json = JSON()
        self._path_separator = path_separator
        self._expand_macros = not nomacros
        self.secrets = None
        self.merged_secrets = None
        self.unmerged_secrets = None
        self._config_imports = config_imports
        self._secrets_imports = secrets_imports
        self._noaws = noaws
        # These booleans are effectively immutable; decided on this default/unchangable behavior.
        self._ignore_missing_macro = True
        self._stringize_non_primitive_types = True
        self._loading = True
        self._load(file_or_dictionary)
        self._loading = False

    def merge_secrets(self, secrets: Config) -> Config:
        merged_config_json, merged_secretst, unmerged_secretst = (
            Config._merge_config_and_secrets(self.json, secrets.json, path_separator=self._path_separator))
        merged_config = Config(merged_config_json, noaws=self._noaws)
        merged_config._file = self.file
        merged_config.secrets = secrets
        merged_config.merged_secrets = merged_secretst
        merged_config.unmerged_secrets = unmerged_secretst
        return merged_config

    def lookup(self, name: str, config: Optional[dict] = None,
               allow_dictionary: bool = False, aws_secret_context_path: Optional[str] = None,
               noaws: bool = False, raw: bool = False, _search_imports: bool = True) -> Optional[str, dict]:

        def lookup_upwards(name: str, config: dict) -> Optional[str]:  # noqa
            nonlocal self
            if parent := config.parent:
                while parent:
                    if ((value := parent.get(name)) is not None) and Config._is_primitive_type(value):
                        return value
                    parent = parent.parent
            return None

        if config is None:
            config = self._json
        value = None
        for index, name_component in enumerate(name_components := name.split(self._path_separator)):
            if value is not None:
                return None  # TODO: Why was this here and what could it mean?
            if not (name_component := name_component.strip()):
                continue
            if not (value := config.get(name_component)):
                # If this is not called during macro expansion (i.e. rather during lookup), and if this
                # is that last name_component, then look straight upwards/outwards in tree for a resolution.
                if (not self._loading) and (index == (len(name_components) - 1)):
                    if (value := lookup_upwards(name_component, config)) is not None:
                        if Config._contains_macro(value):
                            # And if the value contains a macro try resolving from this context.
                            if macro_expanded_value := self._expand_macro_value(value, config):
                                # TODO: Maybe ore tricky stuff needed here.
                                if self._contains_aws_secret_macro(value):
                                    return self._expand_aws_secret_macros(
                                               macro_expanded_value,
                                               aws_secret_context_path=aws_secret_context_path or name, noaws=noaws)
                                return macro_expanded_value
                        return value
                if _search_imports and (docs := self.imports):
                    for doc in docs:
                        if (value := self.lookup(name, doc, _search_imports=False)) is not None:
                            if isinstance(value, dict) and (not allow_dictionary):
                                return None
                            return value
                return None
            if isinstance(value, dict):
                config = value
                value = None
        if value is not None:
            if not self._loading:
                if self._contains_aws_secret_macro(value):
                    return self._expand_aws_secret_macros(value,
                                                          aws_secret_context_path=aws_secret_context_path or name,
                                                          noaws=noaws)
            return value
        elif config and allow_dictionary:
            return config if raw else self._cleanjson(config)
        return None

    @property
    def imports(self) -> List[dict]:
        imports = []
        for key in self._json:
            if (key.startswith(Config._IMPORTED_CONFIG_KEY_PREFIX) or
                key.startswith(Config._IMPORTED_SECRETS_KEY_PREFIX)) and isinstance(self._json[key], dict):  # noqa
                imports.append(self._json[key])
        return imports

    @property
    def secrets_imports(self) -> List[dict]:
        imports = []
        for key in self._json:
            if key.startswith(Config._IMPORTED_SECRETS_KEY_PREFIX) and isinstance(self._json[key], dict):  # noqa
                imports.append(self._json[key])
        return imports

    def contains(self, name: str) -> bool:
        return self.lookup(name, noaws=True) is not None

    @property
    def file(self) -> dict:
        return self._file

    @property
    def json(self) -> dict:
        return self._cleanjson(self._json)

    @property
    def rawjson(self) -> dict:
        return self._json

    def _load(self, file_or_dictionary: Union[str, dict]) -> None:
        if isinstance(file_or_dictionary, dict):
            self._json = JSON(file_or_dictionary)
            self._file = None
        else:
            self._file = file_or_dictionary
            with io.open(file_or_dictionary, "r") as f:
                if self._file.endswith(".yaml") or self._file.endswith(".yml"):
                    self._json = JSON(yaml.safe_load(f))
                else:
                    self._json = JSON(json.load(f))
        for config_import in self._config_imports:
            self._import_config(config_import)
        for secrets_import in self._secrets_imports:
            self._import_secrets(secrets_import)
        if self._expand_macros:
            _ = self._macro_expand(self._json)

    @staticmethod
    def _merge_config_and_secrets(config: JSON, secrets: JSON,
                                  path_separator: str = DEFAULT_PATH_SEPARATOR) -> Tuple[dict, list, list]:
        if not (isinstance(config, dict) or isinstance(secrets, dict)):
            return None, None
        merged = deepcopy(config) ; merged_secrets = [] ; unmerged_secrets = []  # noqa
        def merge(config: dict, secrets: dict, path: str = "") -> None:  # noqa
            nonlocal unmerged_secrets, path_separator
            for key, value in secrets.items():
                key_path = f"{path}{path_separator}{key}" if path else key
                if key not in config:
                    config[key] = secrets[key]
                    merged_secrets.append(key_path)
                elif isinstance(config[key], dict) and isinstance(secrets[key], dict):
                    merge(config[key], secrets[key], path=key_path)
                else:
                    unmerged_secrets.append(key_path)
        merge(merged, secrets)
        return merged, merged_secrets, unmerged_secrets

    def _macro_expand(self, data: dict) -> dict:
        for name in data:  # TODO: check for duplicate keys.
            if not (name := name.strip()):
                continue
            if isinstance(data[name], dict):
                data[name] = self._macro_expand(data[name])
            elif not Config._is_primitive_type(data[name]) and not self._stringize_non_primitive_types:
                raise Exception(f"Non-primitive type found: {name}")
            else:
                data[name] = self._expand_macro_value(str(data[name]), data)

        return data

    def _lookup_macro_value(self, macro_name: str, data: dict) -> Optional[str]:
        if (macro_value := self.lookup(macro_name, data)) is not None:
            if Config._is_primitive_type(macro_value):
                return str(macro_value)
            return None
        data = data.parent
        while data:
            if (macro_value := self.lookup(macro_name, data)) is not None:
                if Config._is_primitive_type(macro_value):
                    return str(macro_value)
                return None
            data = data.parent
        return None

    def _expand_macro_value(self, value: str, data: dict) -> Optional[str]:
        expanding_macros = set()
        missing_macro_found = False
        original_simple_macros_to_retain = {}
        while True:
            if not (match := Config._MACRO_PATTERN.search(value)):
                break
            if (macro_name := match.group(1)) and (macro_value := self._lookup_macro_value(macro_name, data)):
                if Config._is_macro(macro_value) and (not self._is_aws_secret_macro(macro_value)):
                    original_simple_macros_to_retain[macro_value] = macro_name
                #
                # Note the _is_macro call above is a bit of a special case; if the macro we are expanding resolves
                # simply to another macro, then retain the original macro; this can be useful for this for example:
                #
                # foursight:
                #   SSH_TUNNEL_ES_NAME_PREFIX: ssh-tunnel-es-proxy
                #   SSH_TUNNEL_ES_NAME: ${SSH_TUNNEL_ES_NAME_PREFIX}-${SSH_TUNNEL_ES_ENV}-${SSH_TUNNEL_ES_PORT}
                #   SSH_TUNNEL_ES_ENV: ${AWS_PROFILE}
                #   4dn:
                #     AWS_PROFILE: 4dn
                #     SSH_TUNNEL_ES_ENV: ${AWS_PROFILE}-mastertest
                #     SSH_TUNNEL_ES_PORT: 9203
                #     dev:
                #       IDENTITY: whatever
                #   smaht:
                #     wolf:
                #       AWS_PROFILE: smaht-wolf
                #       SSH_TUNNEL_ES_PORT: 9209
                #       IDENTITY: whatever
                #
                # For lookup of foursight/smaht/wolf/SSH_TUNNEL_ES_NAME we will get ssh-tunnel-es-proxy-smaht-wolf-9209
                # as it takes the default foursight/SSH_TUNNEL_ES_ENV value which is foursight/smaht/wolf/AWS_PROFILE,
                # i.e. smaht-wolf; remember - we evaluate/expand the macro starting in the context of the lookup, i.e.
                # foursight/smaht/wolf. But for lookup of foursight/4dn/dev/SSH_TUNNEL_ES_NAME we want to (and do)
                # get ssh-tunnel-es-proxy-4dn-mastertest-9203 because retaining the ${SSH_TUNNEL_ES_ENV} in
                # foursight/SSH_TUNNEL_ES_NAME we will pick up the "overriding" foursight/4dn/SSH_TUNNEL_ES_ENV;
                # without this _is_macro call, we would get ssh-tunnel-es-proxy-4dn-9203 because ${SSH_TUNNEL_ES_ENV}
                # in foursight/SSH_TUNNEL_ES_NAME would have been expaned to ${AWS_PROFILE}.
                #
                if macro_name in expanding_macros:
                    raise Exception(f"Circular macro definition found: {macro_name}")
                expanding_macros.add(macro_name)
                value = value.replace(f"${{{macro_name}}}", macro_value)
            elif self._ignore_missing_macro:
                missing_macro_found = True
                value = value.replace(f"{Config._MACRO_START}{macro_name}{Config._MACRO_END}",
                                      f"{Config._MACRO_HIDE_START}{macro_name}{Config._MACRO_HIDE_END}")
            else:
                raise Exception(f"Macro name not found: {macro_name}")
        if missing_macro_found and self._ignore_missing_macro:
            value = value.replace(Config._MACRO_HIDE_START, Config._MACRO_START)
            value = value.replace(Config._MACRO_HIDE_END, Config._MACRO_END)
        if original_simple_macros_to_retain and self._contains_macro(value):
            for simple_macro in original_simple_macros_to_retain:
                original_simple_macro = original_simple_macros_to_retain[simple_macro]
                value = value.replace(simple_macro, f"{Config._MACRO_START}{original_simple_macro}{Config._MACRO_END}")
        return value

    @staticmethod
    def _is_macro(value: str) -> bool:
        return value.startswith(Config._MACRO_START) and value.endswith(Config._MACRO_END)

    @staticmethod
    def _contains_macro(value: str) -> bool:
        if (index := value.find(Config._MACRO_START)) >= 0:
            if value[index + len(Config._MACRO_START):].find(Config._MACRO_END) > 0:
                return True
        return False

    @staticmethod
    def _is_aws_secret_macro_name(macro_name: str) -> bool:
        return macro_name.startswith(Config._AWS_SECRET_MACRO_NAME_PREFIX)

    @staticmethod
    def _is_aws_secret_macro(macro: str) -> bool:
        return (macro.startswith(Config._AWS_SECRET_MACRO_START) and
                macro.endswith(Config._AWS_SECRET_MACRO_END))

    @staticmethod
    def _contains_aws_secret_macro(value: str) -> bool:
        if (index := value.find(Config._AWS_SECRET_MACRO_START)) >= 0:
            if value[index + len(Config._AWS_SECRET_MACRO_START):].find(Config._AWS_SECRET_MACRO_END) > 0:
                return True
        return False

    def _expand_aws_secret_macros(self, value: str, aws_secret_context_path: str, noaws: bool = False) -> Optional[str]:
        while ((match := Config._AWS_SECRET_MACRO_PATTERN.search(value)) and
               (secret_specifier := match.group(1))):
            if macro_value := self._expand_aws_secret_macro(secret_specifier, aws_secret_context_path, noaws=noaws):
                macro = f"{Config._AWS_SECRET_MACRO_START}{secret_specifier}{Config._AWS_SECRET_MACRO_END}"
                value = value.replace(macro, macro_value)
            else:
                # Hide the macro temporarily for this loop.
                value = value.replace(f"{Config._AWS_SECRET_MACRO_START}{secret_specifier}{Config._MACRO_END}",
                                      f"{Config._MACRO_HIDE_START}{secret_specifier}{Config._MACRO_HIDE_END}")
        # Unhide unfound macros.
        value = value.replace(Config._MACRO_HIDE_START, Config._AWS_SECRET_MACRO_START)
        value = value.replace(Config._MACRO_HIDE_END, Config._MACRO_END)
        return value

    def _expand_aws_secret_macro(self, secret_specifier: str,
                                 aws_secret_context_path: str, noaws: bool = False) -> Optional[str]:
        if (index := secret_specifier.find(self._path_separator)) > 0:
            secret_name = secret_specifier[index + 1:]
            secrets_name = secret_specifier[0:index]
        else:
            secret_name = secret_specifier
            secrets_name = None
            if (index := aws_secret_context_path.rfind(self._path_separator)) > 0:
                aws_secret_context_path = aws_secret_context_path[:index]
                secrets_name_path = f"{aws_secret_context_path}/{Config._AWS_SECRET_NAME_NAME}"
                secrets_name = self.lookup(secrets_name_path)
        if secrets_name and secret_name:
            if secret_value := self._lookup_aws_secret(secrets_name, secret_name, noaws=noaws):
                return secret_value
        return None

    @lru_cache(maxsize=64)
    def _lookup_aws_secret(self, secrets_name: str, secret_name: str, noaws: bool = False) -> Optional[str]:
        if noaws or self._noaws:
            return "__noaws__"
        try:
            if secrets := self._aws_get_secret_value(secrets_name):
                return secrets.get(secret_name)
        except Exception:
            global SUPPRESS_AWS_SECRET_NOT_FOUND_WARNING
            if not SUPPRESS_AWS_SECRET_NOT_FOUND_WARNING:
                warning(f"Cannot find AWS secret: {secrets_name}/{secret_name}")
            return None

    def _aws_get_secret_value(self, secrets_name: str) -> Optional[dict]:
        boto_secrets = BotoClient("secretsmanager")
        if secrets := boto_secrets.get_secret_value(SecretId=secrets_name):
            return json.loads(secrets.get("SecretString"))

    def _import_config(self, config: dict, name: Optional[str] = None) -> None:
        if isinstance(config, dict) and config:
            self._json[f"{Config._IMPORTED_CONFIG_KEY_PREFIX}{name or ''}"] = config

    def _import_secrets(self, secrets: dict, name: Optional[str] = None) -> None:
        if isinstance(secrets, dict) and secrets:
            self._json[f"{Config._IMPORTED_SECRETS_KEY_PREFIX}{name or ''}"] = secrets

    @staticmethod
    def _is_primitive_type(value: Any) -> bool:  # noqa
        return isinstance(value, (int, float, str, bool))

    @staticmethod
    def _cleanjson(data: dict) -> dict:
        data = deepcopy(data)
        def remove_imported_configs(data: dict) -> None:  # noqa
            todelete = []
            for key in data:
                if (key.startswith(Config._IMPORTED_CONFIG_KEY_PREFIX) or
                    key.startswith(Config._IMPORTED_SECRETS_KEY_PREFIX)):  # noqa
                    todelete.append(key)
            for key in todelete:
                del data[key]
        remove_imported_configs(data)
        return data


def warning(message: str) -> None:
    print(f"WARNING: {message}", file=sys.stderr, flush=True)


def error(message: str, usage: bool = False, status: int = 1,
          exception: Optional[Exception] = None, trace: bool = False) -> None:
    print(f"ERROR: {message}", file=sys.stderr)
    if usage:
        _usage()
    if isinstance(exception, Exception):
        print(str(exception))
    if trace:
        traceback.print_exc()
    sys.exit(status)


def _usage():
    print(f"{chars.rarrow} hmx-config reads named value from {DEFAULT_CONFIG_FILE_NAME} or"
          f" {DEFAULT_SECRETS_FILE_NAME} in: {DEFAULT_CONFIG_DIR}")
    print(f"  {chars.rarrow_hollow} usage: python hms_config.py"
          f" [ path/name [-json] | [-nocolor | -nomerge | -nosort | -json | -yaml | -show] ]")
    sys.exit(1)


if __name__ == "__main__":
    main()



import io
import json
import os
import sys
from typing import Tuple
import uuid
from dcicutils.command_utils import yes_or_no
from hms_utils.argv import ARGV
from hms_utils.chars import chars
from hms_utils.portal.portal_utils import Portal as Portal
from hms_utils.type_utils import is_uuid

_ACCESS_KEY_TYPE_NAME = "AccessKey"
_DEFAULT_APP = "smaht"
_DEFAULT_SERVER = "http://localhost:8000"


def main():

    argv = ARGV({
        ARGV.OPTIONAL(str): ["--env"],
        ARGV.OPTIONAL(str): ["--ini"],
        ARGV.OPTIONAL(str): ["--app"],
        ARGV.REQUIRED(str): ["--user"],
        ARGV.OPTIONAL(str): ["--access-keys-file-property-name", "--name"],
        ARGV.OPTIONAL(bool): ["--update"],
        ARGV.OPTIONAL(bool): ["--update-database"],
        ARGV.OPTIONAL(bool): ["--update-keys", "--update-keys-file"],
        ARGV.OPTIONAL(str): ["--server"],
        ARGV.OPTIONAL(bool): ["--quiet"],
        ARGV.OPTIONAL(bool): ["--verbose"],
        ARGV.OPTIONAL(bool): ["--debug"],
        ARGV.OPTIONAL(bool): ["--yes", "--force"],
        ARGV.OPTIONAL(bool): ["--ping"],
        ARGV.OPTIONAL(bool): ["--version"]
    })

    portal = Portal.create(argv.ini or argv.env, app=argv.app, verbose=argv.verbose, debug=argv.debug, ping=argv.ping)

    if not (isinstance(user := portal.get_metadata(f"/users/{argv.user}", raise_exception=False), dict) and
            is_uuid(user_uuid := user.get("uuid"))):
        print(f"Cannot find user: {argv.user}")
        exit(0)

    access_keys_file = portal.keys_file
    access_keys_file_property_name = argv.access_keys_file_property_name or portal.env
    access_keys_server = portal.server or argv.server or _DEFAULT_SERVER
    access_keys_app = portal.app or argv.app or _DEFAULT_APP

    if not access_keys_file:
        access_keys_file = os.path.expanduser(f"~/.{access_keys_app}-keys.json")
    if not access_keys_file_property_name:
        access_keys_file_property_name = f"{access_keys_app}-local"

    access_key_id, access_key_secret, access_key_secret_hash = _generate_access_key_elements()
    access_keys_file_item = _generate_access_keys_file_item(access_key_id, access_key_secret, access_keys_server)
    access_key_portal_inserts_item = _generate_access_key_portal_inserts_item(access_key_id,
                                                                              access_key_secret_hash, user_uuid)
    if not argv.quiet:
        print(f"Newly generated portal access-key record suitable for: {access_keys_file} ...")
        print(json.dumps(access_keys_file_item, indent=4))
        print(f"Newly generated portal access-key record suitable for insert ...")
        print(json.dumps(access_key_portal_inserts_item, indent=4))

    if argv.update:
        argv.update_database = True
        argv.update_keys = True

    if argv.update_database:
        if (not argv.quiet) or argv.verbose:
            print(f"Updating portal database with newly generated access-key item:"
                  f" {access_keys_server} {chars.dot} {access_key_id}")
        _update_portal_with_access_key(portal, access_key_portal_inserts_item)

    if argv.update_keys:
        if (not argv.quiet) or argv.verbose:
            print(f"Updating access-keys file with newly generated access-key item:"
                  f" {access_keys_file} {chars.dot} {access_keys_file_property_name}")
        if os.path.exists(access_keys_file):
            try:
                with io.open(access_keys_file, "r") as f:
                    access_keys_file_json = json.load(f)
            except Exception:
                print(f"Cannot read existing access-keys file: {access_keys_file}")
                exit(1)
            if access_keys_file_json.get(access_keys_file_property_name):
                if not argv.yes:
                    print(f"Item already exists for access-key item ({access_keys_file_property_name})"
                          f" in access-keys file: {access_keys_file}")
                    if not yes_or_no(f"Overwrite access-key item {access_keys_file_property_name}?"):
                        return 1
                elif argv.verbose:
                    print(f"Overwriting existing item for access-key ({access_keys_file_property_name})"
                          f" in access-keys file: {access_keys_file}")
            access_keys_file_json[access_keys_file_property_name] = access_keys_file_item
        else:
            access_keys_file_json = access_keys_file_item
        try:
            with io.open(access_keys_file, "w") as f:
                json.dump(access_keys_file_json, fp=f, indent=4)
        except Exception:
            print(f"Cannot write access-keys file: {access_keys_file}")
            exit(1)


def _generate_access_key_portal_inserts_item(access_key_id: str, access_key_secret_hash: str, user_uuid: str) -> dict:
    return {
        "status": "current",
        "user": user_uuid,
        "description": f"Manually generated local access-key for testing.",
        "access_key_id": access_key_id,
        "secret_access_key_hash": access_key_secret_hash,
        "uuid": str(uuid.uuid4())
    }


def _generate_access_keys_file_item(access_key_id: str, access_key_secret: str, server: str) -> dict:
    return {
        "key": access_key_id,
        "secret": access_key_secret,
        "server": server
    }


def _generate_access_key_elements() -> Tuple[str, str, str]:
    from passlib.context import CryptContext
    from passlib.registry import register_crypt_handler
    try:
        from snovault.authentication import (
            generate_password as snovault_generate_access_key_secret,
            generate_user as snovault_generate_access_key
        )
        from snovault.edw_hash import EDWHash as snovault_edwhash
    except Exception:
        print("Cannot import dcicsnovault modules which are needed for access-key secret generation.")
        exit(1)
    # Generate access-key and secret like snovault does.
    access_key = snovault_generate_access_key()
    access_key_secret = snovault_generate_access_key_secret()
    # Generate access-key secret hash like snovault does. See: snovault/authentication.py
    passlib_properties = {"schemes": "edw_hash, unix_disabled"}
    register_crypt_handler(snovault_edwhash)
    access_key_hash_secret = CryptContext(**passlib_properties).hash(access_key_secret)
    # Return the access-key and its associated secret and secret hash (for the portal database item) values.
    return access_key, access_key_secret, access_key_hash_secret


def _update_portal_with_access_key(portal: Portal, data: dict) -> bool:
    if portal.ini_file:
        return _load_data(portal, data)
    else:
        return portal.post_metadata(_ACCESS_KEY_TYPE_NAME, data)
    pass


def _load_data(portal: Portal, data: dict, data_type: str = _ACCESS_KEY_TYPE_NAME) -> bool:
    try:
        from snovault.loadxl import load_all as snovault_load_all
    except Exception:
        print("Cannot import dcicsnovault modules which are needed for local access-key loading.")
        exit(1)
    if isinstance(data, dict):
        data = [data]
    elif not isinstance(data, list):
        return False
    if not isinstance(data_type, str):
        return False
    data = {data_type: data}
    snovault_load_all(portal.vapp, inserts=data, docsdir=None, overwrite=True, itype=[data_type], from_json=True)
    return True


if __name__ == "__main__":
    sys.exit(main())



import re
import requests
import sys
from hms_utils.chars import chars


def parse_database_count(value):
    match = re.search(r"DB:\s*(\d+)\s+ES:\s*(\d+)", value)
    if match:
        count = int(match.group(1))  # First captured group is the DB count
        return count
    return None


def parse_elasticsearch_count(value):
    match = re.search(r"DB:\s*(\d+)\s+ES:\s*(\d+)", value)
    if match:
        count = int(match.group(2))
        return count
    return None


def report_indexing_status_and_counts(name, domain, verbose):

    print()
    print(f"{chars.rarrow} {name}: {domain}")
    url = f"https://{domain}/indexing_status?format=json"
    response = requests.get(url)
    response = response.json()
    print(f"  {chars.rarrow_hollow} INDEX STATUS: {'' if response['status'] == 'Success' else response['status']}")
    print(f"    - Primary   waiting:   {response['primary_waiting']}")
    print(f"    - Primary   inflight:  {response['primary_inflight']}")
    print(f"    - Secondary waiting:   {response['secondary_waiting']}")
    print(f"    - Secondary inflight:  {response['secondary_inflight']}")
    print(f"    - DLQ       waiting:   {response['dlq_waiting']}")
    print(f"    - DLQ       inflight:  {response['dlq_inflight']}")

    url = f"https://{domain}/counts?format=json"
    response = requests.get(url)
    response = response.json()
    print(f"  {chars.rarrow_hollow} COUNTS:")
    database_elasticsearch_counts = response['db_es_total']
    database_count = parse_database_count(database_elasticsearch_counts)
    elasticsearch_count = parse_elasticsearch_count(database_elasticsearch_counts)
    if database_count > elasticsearch_count:
        database_suffix = f" {chars.rarrow_hollow} {database_count - elasticsearch_count} more"
        elasticsearch_suffix = ""
    elif elasticsearch_count > database_count:
        database_suffix = ""
        elasticsearch_suffix = f" {chars.rarrow_hollow} {elasticsearch_count - database_count} more"
    else:
        database_suffix = ""
        elasticsearch_suffix = ""
    print(f"    - Database:      {database_count}"
          f" {chars.check if database_count == elasticsearch_count else chars.xmark}{database_suffix}")
    print(f"    - ElasticSearch: {elasticsearch_count}"
          f" {chars.check if database_count == elasticsearch_count else chars.xmark}{elasticsearch_suffix}")
    if verbose and (database_count != elasticsearch_count):
        for item in response["db_es_compare"]:
            value = response["db_es_compare"][item]
            item_database_count = parse_database_count(value)
            item_elasticsearch_count = parse_elasticsearch_count(value)
            if item_database_count != item_elasticsearch_count:
                if database_count > elasticsearch_count:
                    database_suffix = f" {chars.rarrow_hollow} {database_count - elasticsearch_count} more"
                    elasticsearch_suffix = ""
                elif elasticsearch_count > database_count:
                    database_suffix = ""
                    elasticsearch_suffix = f" {chars.rarrow_hollow} {elasticsearch_count - database_count} more"
                else:
                    database_suffix = ""
                    elasticsearch_suffix = ""
                print(f"      {chars.rarrow_hollow} {item}:")
                print(f"        - Database:      {item_database_count}{database_suffix}")
                print(f"        - ElasticSearch: {item_elasticsearch_count}{elasticsearch_suffix}")


def main():

    smaht = False
    fourdn = False
    verbose = False

    if "smaht" in sys.argv:
        smaht = True
    if "4dn" in sys.argv:
        fourdn = True
    if not (("smaht" in sys.argv) or ("4dn" in sys.argv)):
        smaht = True
    if ("--verbose" in sys.argv) or ("-verbose" in sys.argv):
        verbose = True

    if smaht:
        report_indexing_status_and_counts("SMaHT STAGING", "staging.smaht.org", verbose)
        report_indexing_status_and_counts("SMaHT DATA", "data.smaht.org", verbose)

    if fourdn:
        report_indexing_status_and_counts("4DN STAGING", "staging.4dnucleome.org", verbose)
        report_indexing_status_and_counts("4DN DATA", "data.4dnucleome.org", verbose)

    print()


if __name__ == "__main__":
    main()



from termcolor import colored


def terminal_color(value: str,
                   color: str,
                   dark: bool = False,
                   bold: bool = False,
                   underline: bool = False,
                   nocolor: bool = False) -> str:
    if nocolor is True:
        return value
    attributes = []
    if dark is True:
        attributes.append("dark")
    if bold is True:
        attributes.append("bold")
    if underline is True:
        attributes.append("underline")
    return colored(value, color.lower(), attrs=attributes)
